#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010-2017.

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Esben S. Nielsen <esn@dmi.dk>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""The :mod:`mpop.scene` module defines satellite scenes. They are defined as
generic classes, to be inherited when needed.

A scene is a set of :mod:`mpop.channel` objects for a given time, and sometimes
also for a given area.
"""
import ConfigParser
import copy
import datetime
import imp
import logging
import os.path
import sys
import types
import weakref

import numpy as np

import mpop.satin
from mpop import CONFIG_PATH
from mpop.channel import Channel, NotLoadedError
from mpop.utils import OrderedConfigParser

LOG = logging.getLogger(__name__)

try:
    # Work around for on demand import of pyresample. pyresample depends
    # on scipy.spatial which memory leaks on multiple imports
    is_pyresample_loaded = False
    from pyresample.geometry import AreaDefinition, SwathDefinition
    import mpop.projector
    is_pyresample_loaded = True
except ImportError:
    LOG.warning("pyresample missing. Can only work in satellite projection")


class Satellite(object):

    """This is the satellite class. It contains information on the satellite.
    """

    def __init__(self, (satname, number, variant)=(None, None, None)):
        try:
            self.satname = satname or "" or self.satname
        except AttributeError:
            self.satname = satname or ""
        try:
            self.number = number or "" or self.number
        except AttributeError:
            self.number = number or ""
        try:
            self.variant = variant or "" or self.variant
        except AttributeError:
            self.variant = variant or ""

    @property
    def fullname(self):
        """Full name of the satellite, that is platform name and number
        (eg "metop02").
        """
        return self.variant + self.satname + self.number

    def sat_nr(self, string=False):
        import re
        sat_nr = re.findall(r'\d+', self.fullname)[0]
        if string:
            return sat_nr
        else:
            return int(sat_nr)

    @classmethod
    def remove_attribute(cls, name):
        """Remove an attribute from the class.
        """
        return delattr(cls, name)

    @classmethod
    def add_method(cls, func):
        """Add a method to the class.
        """
        return setattr(cls, func.__name__, func)

    def add_method_to_instance(self, func):
        """Add a method to the instance.
        """
        return setattr(self, func.__name__,
                       types.MethodType(func, self.__class__))


class SatelliteScene(Satellite):

    """This is the satellite scene class. It is a capture of the satellite
    (channels) data at given *time_slot* and *area_id*/*area*.
    """

    def __init__(self, time_slot=None, area_id=None, area=None,
                 orbit=None, satellite=(None, None, None)):

        Satellite.__init__(self, satellite)

        # if(time_slot is not None and
        #   not isinstance(time_slot, datetime.datetime)):
        #    raise TypeError("Time_slot must be a datetime.datetime instance.")

        self.time_slot = time_slot

        self.area_id = None
        self.area_def = None

        if area_id is not None:
            from warnings import warn
            warn("The *area_id* attribute is deprecated."
                 "Please use *area* instead.",
                 DeprecationWarning)
            if not isinstance(area_id, str):
                raise TypeError("Area must be a string.")

        self.area = area_id

        if area is not None:
            self.area = area

        self.orbit = orbit

        self.info = {}

        self.lat = None
        self.lon = None

    def get_area(self):
        """Getter for area.
        """
        return self.area_def or self.area_id

    def set_area(self, area):
        """Setter for area.
        """
        if area is None:
            self.area_def = None
            self.area_id = None
        elif isinstance(area, str):
            self.area_id = area
            self.area_def = None
        else:
            try:
                dummy = area.area_extent
                dummy = area.x_size
                dummy = area.y_size
                dummy = area.proj_id
                dummy = area.proj_dict
                self.area_def = area
                self.area_id = None
            except AttributeError:
                try:
                    dummy = area.lons
                    dummy = area.lats
                    self.area_def = area
                    self.area_id = None
                except AttributeError:
                    raise TypeError(("Malformed area argument. "
                                     "Should be a string or an area object. "
                                     "Not %s") % type(area))

    area = property(get_area, set_area)


class SatelliteInstrumentScene(SatelliteScene):

    """This is the satellite instrument class. It is an abstract channel
    container, from which all concrete satellite scenes should be derived.

    The constructor accepts as optional arguments the *time_slot* of the scene,
    the *area* on which the scene is defined (this can be used for slicing of
    big datasets, or can be set automatically when loading), and *orbit* which
    is a string giving the orbit number.
    """
    channel_list = []

    def __init__(self, time_slot=None, area_id=None, area=None,
                 orbit=None, satellite=(None, None, None), instrument=None):

        SatelliteScene.__init__(self, time_slot, area_id, area,
                                orbit, satellite)

        try:
            self.instrument_name = instrument or self.instrument_name
        except AttributeError:
            self.instrument_name = None

        self.channels = []

        self.end_time = None
        if isinstance(self.time_slot, (tuple, list)):
            self.time_slot, self.end_time = self.time_slot

        try:
            conf = OrderedConfigParser()
            conf.read(os.path.join(CONFIG_PATH, self.fullname + ".cfg"))

            for section in conf.sections():
                if(not section[:-1].endswith("level") and
                   not section.endswith("granules") and
                   section.startswith(self.instrument_name)):
                    name = eval(conf.get(section, "name"))
                    try:
                        w_range = eval(conf.get(section, "frequency"))
                    except ConfigParser.NoOptionError:
                        w_range = (-np.inf, -np.inf, -np.inf)
                    try:
                        resolution = eval(conf.get(section, "resolution"))
                    except ConfigParser.NoOptionError:
                        resolution = 0
                    self.channels.append(Channel(name=name,
                                                 wavelength_range=w_range,
                                                 resolution=resolution))

        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            for name, w_range, resolution in self.channel_list:
                self.channels.append(Channel(name=name,
                                             wavelength_range=w_range,
                                             resolution=resolution))

        self.channels_to_load = set([])

    def __getitem__(self, key, aslist=False):
        if isinstance(key, float):
            channels = [chn for chn in self.channels
                        if(hasattr(chn, "wavelength_range") and
                           chn.wavelength_range[0] <= key and
                           chn.wavelength_range[2] >= key)]
            channels = sorted(channels,
                              lambda ch1, ch2:
                                  ch1.__cmp__(ch2, key))

        elif isinstance(key, str):
            channels = [chn for chn in self.channels
                        if chn.name == key]
            channels = sorted(channels)

        elif isinstance(key, int):
            channels = [chn for chn in self.channels
                        if int(np.round(chn.resolution)) == key]
            channels = sorted(channels)

        elif isinstance(key, (tuple, list)):
            if len(key) == 0:
                raise KeyError("Key list must contain at least one element.")
            channels = self.__getitem__(key[0], aslist=True)
            if len(key) > 1 and len(channels) > 0:
                dummy_instance = SatelliteInstrumentScene()
                dummy_instance.channels = channels
                channels = dummy_instance.__getitem__(key[1:], aslist=True)
        else:
            raise TypeError("Malformed key: " + str(key))

        if len(channels) == 0:
            raise KeyError("No channel corresponding to " + str(key) + ".")
        elif aslist:
            return channels
        else:
            return channels[0]

    def __delitem__(self, key):
        """Delete a channel.
        """
        chan = self[key]
        self.channels.remove(chan)
        del chan

    def __setitem__(self, key, data):
        # Add a channel if it is not already in the scene. Works only if key is
        # a string.
        try:
            if key not in self:
                # if it's a blob with name and data, add it as is.
                if hasattr(data, "name") and hasattr(data, "data"):
                    self.channels.append(data)
                else:
                    kwargs = {"name": key}
                    for attr in ["wavelength_range", "resolution"]:
                        try:
                            kwargs[attr] = getattr(data, attr)
                        except (AttributeError, NameError):
                            pass
                    self.channels.append(Channel(**kwargs))
        except AttributeError:
            pass

        # Add the data.
        if isinstance(data, np.ma.core.MaskedArray):
            self[key].data = data
        else:
            try:
                self[key].data = data.data
            except AttributeError:
                self[key].data = data

        # if isinstance(data, Channel):
        #     self.channels.append(Channel(name=key,
        #                              wavelength_range=data.wavelength_range,
        #                              resolution=data.resolution))
        #     self[key].data = data.data
        # else:
        #     try:
        #         self[key].data = data
        #     except KeyError:
        #         self.channels.append(Channel(name=key))
        #         self[key].data = data

    def __str__(self):
        return "\n".join([str(chn) for chn in self.channels])

    def __iter__(self):
        return self.channels.__iter__()

    def _set_reader(self, pformat):
        """Gets the reader for *pformat* format, and puts it in the `reader`
        attribute.
        """

        elements = pformat.split(".")
        if len(elements) == 1:
            reader_module = pformat
            reading_element = 'load'

            reader = "mpop.satin." + reader_module

            # Loading old style plugins
            reader_module = pformat
            LOG.info("old style plugin: " + pformat)

            try:
                # Look for builtin reader
                imp.find_module(reader_module, mpop.satin.__path__)
            except ImportError:
                # Look for custom reader
                loader = __import__(reader_module, globals(),
                                    locals(), [reading_element])
            else:
                loader = __import__(reader, globals(),
                                    locals(), [reading_element])

            # Build a custom Reader plugin on the fly...
            from mpop.plugin_base import Reader
            reader_class = type(elements[-1].capitalize() + "Reader",
                                (Reader,),
                                {"pformat": elements[-1]})

            reader_instance = reader_class(self)

            # ... and set its "load" attribute with the "load" function of the
            # loader module
            loader = getattr(loader, reading_element)
            setattr(reader_instance, reading_element, loader)

            setattr(self, elements[-1] + "_reader", reader_instance)

        else:
            reader_module = ".".join(elements[:-1])
            reading_element = elements[-1]

            reader = "mpop.satin." + reader_module
            try:
                # Look for builtin reader
                imp.find_module(reader_module, mpop.satin.__path__)
            except ImportError:
                # Look for custom reader
                loader = __import__(reader_module, globals(),
                                    locals(), [reading_element])
            else:
                loader = __import__(reader, globals(),
                                    locals(), [reading_element])

            loader = getattr(loader, reading_element)
            reader_instance = loader(self)
            setattr(self, loader.pformat + "_reader", reader_instance)

        return reader_instance

    def load(self, channels=None, load_again=False, area_extent=None, **kwargs):
        """Load instrument data into the *channels*. *Channels* is a list or a
        tuple containing channels we will load data into, designated by there
        center wavelength (float), resolution (integer) or name (string). If
        None, all channels are loaded.

        The *load_again* boolean flag allows to reload the channels even they
        have already been loaded, to mirror changes on disk for example. This
        is false by default.

        The *area_extent* keyword lets you specify which part of the data to
        load. Given as a 4-element sequence, it defines the area extent to load
        in satellite projection.

        The other keyword arguments are passed as is to the reader
        plugin. Check the corresponding documentation for more details.
        """

        # Set up the list of channels to load.
        if channels is None:
            for chn in self.channels:
                self.channels_to_load |= set([chn.name])

        elif isinstance(channels, (list, tuple, set)):
            self.channels_to_load = set()
            for chn in channels:
                try:
                    self.channels_to_load |= set([self[chn].name])
                except KeyError:
                    self.channels_to_load |= set([chn])

        else:
            raise TypeError("Channels must be a list/"
                            "tuple/set of channel keys!")

        loaded_channels = [chn.name for chn in self.loaded_channels()]
        if load_again:
            for chn in self.channels_to_load:
                if chn in loaded_channels:
                    self.unload(chn)
                    loaded_channels = []
        else:
            for chn in loaded_channels:
                self.channels_to_load -= set([chn])

        # find the plugin to use from the config file
        conf = ConfigParser.ConfigParser()
        try:
            conf.read(os.path.join(CONFIG_PATH, self.fullname + ".cfg"))
            if len(conf.sections()) == 0:
                raise ConfigParser.NoSectionError(("Config file did "
                                                   "not make sense"))
            levels = [section for section in conf.sections()
                      if section.startswith(self.instrument_name + "-level")]
        except ConfigParser.NoSectionError:
            LOG.warning("Can't load data, no config file for " + self.fullname)
            self.channels_to_load = set()
            return

        levels.sort()

        if levels[0] == self.instrument_name + "-level1":
            levels = levels[1:]

        if len(levels) == 0:
            raise ConfigParser.NoSectionError(
                self.instrument_name + "-levelN (N>1) to tell me how to" +
                " read data... Not reading anything.")

        for level in levels:
            if len(self.channels_to_load) == 0:
                return

            if "reader_level" in kwargs.keys():
                if kwargs["reader_level"] != None:
                    LOG.debug(
                        "Using explecit definition of reader level: " + kwargs["reader_level"])
                    if kwargs["reader_level"] != level:
                        continue

            LOG.debug("Looking for sources in section " + level)
            reader_name = conf.get(level, 'format')
            try:
                reader_name = eval(reader_name)
            except NameError:
                reader_name = str(reader_name)
            LOG.debug("Using plugin mpop.satin." + reader_name)

            # read the data
            reader = "mpop.satin." + reader_name
            try:
                reader_instance = self._set_reader(reader_name)
                if area_extent is not None:
                    if(isinstance(area_extent, (tuple, list)) and
                       len(area_extent) == 4):
                        kwargs["area_extent"] = area_extent
                    else:
                        raise ValueError("Area extent must be a sequence of "
                                         "four numbers.")

                reader_instance.load(self, **kwargs)
            except ImportError, err:
                LOG.exception("ImportError while loading " + reader_name + ": "
                              + str(err))
                continue
            loaded_channels = set([chn.name for chn in self.loaded_channels()])
            just_loaded = loaded_channels & self.channels_to_load
            if len(just_loaded) == 0:
                LOG.info("No channels loaded with " + reader + ".")
            self.channels_to_load -= loaded_channels
            LOG.debug("Successfully loaded: " + str(just_loaded))

        if len(self.channels_to_load) > 0:
            LOG.warning("Unable to import channels "
                        + str(self.channels_to_load))

        self.channels_to_load = set()

    def save(self, filename, to_format="netcdf4", **options):
        """Saves the current scene into a file of format *to_format*. Supported
        formats are:

        - *netcdf4*: NetCDF4 with CF conventions.
        """

        writer = "satout." + to_format
        try:
            writer_module = __import__(writer, globals(), locals(), ["save"])
        except ImportError, err:
            raise ImportError("Cannot load " + writer + " writer: " + str(err))

        return writer_module.save(self, filename, **options)

    def unload(self, *channels):
        """Unloads *channels* from
        memory. :meth:`mpop.scene.SatelliteInstrumentScene.load` must be called
        again to reload the data.
        """
        for chn in channels:
            try:
                self[chn].data = None
            except AttributeError:
                LOG.warning("Can't unload channel" + str(chn))

    def add_to_history(self, message):
        """Adds a message to history info.
        """
        timestr = datetime.datetime.utcnow().isoformat()
        timed_message = str(timestr + " - " + message)
        if not self.info.get("history", ""):
            self.info["history"] = timed_message
        else:
            self.info["history"] += "\n" + timed_message

    def check_channels(self, *channels):
        """Check if the *channels* are loaded, raise an error otherwise.
        """
        for chan in channels:
            if not self[chan].is_loaded():
                raise NotLoadedError("Required channel %s not loaded,"
                                     " aborting." % chan)

        return True

    def loaded_channels(self):
        """Return the set of loaded_channels.
        """
        return set([chan for chan in self.channels if chan.is_loaded()])

    def get_orbital(self):
        from pyorbital.orbital import Orbital
        from pyorbital import tlefile

        from pyorbital.tlefile import get_norad_line
        sat_line = get_norad_line(self.satname, self.number)
        self.orbital = Orbital(sat_line)

        return self.orbital

    def estimate_cth(self, cth_atm="best", time_slot=None):
        """
        General purpose
        ===============
           Estimation of the cloud top height using the 10.8 micron channel
           limitations: this is the most simple approach
           a simple fit of the ir108 to the temperature profile
                * no correction for water vapour or any other trace gas
                * no viewing angle dependency
                * no correction for semi-transparent clouds
                * no special treatment of temperature inversions
        Example call
        ============
           data.estimate_cth(cth_atm="best")
        input arguments
        ===============
          cth_atm    * using temperature profile to estimate the cloud top height
                       possible choices are (see estimate_cth in mpop/tools.py):
                       "standard", "tropics", "midlatitude summer", "midlatitude winter", "subarctic summer", "subarctic winter"
                       this will choose the corresponding atmospheric AFGL temperature profile
                     * new choice: "best" -> choose according to central (lon,lat) and time from:
                       "tropics", "midlatitude summer", "midlatitude winter", "subarctic summer", "subarctic winter"
          time_slot  current observation time as (datetime.datetime() object)
                     time_slot option can be omitted, the function tries to use self.time_slot
        """

        print "*** Simple estimation of Cloud Top Height with IR_108 channel"

        # check if IR_108 is loaded
        loaded_channels = [chn.name for chn in self.loaded_channels()]
        if "IR_108" not in loaded_channels:
            print "*** Error in estimate_cth (mpop/scene.py)"
            print "    IR_108 is required to estimate CTH, but not loaded"
            quit()
        else:
            ir108 = self["IR_108"].data

        # choose atmosphere
        if cth_atm.lower() == "best":
            # get central lon/lat coordinates
            (yc, xc) = ir108.shape
            (lon, lat) = self.area.get_lonlat(yc / 2, xc / 2)

            if time_slot == None:
                if hasattr(self, 'time_slot'):
                    time_slot = self.time_slot
                else:
                    print "*** Error, in estimate_cth (mpop/channel.py)"
                    print "    when using cth_atm=\"best\" also the time_slot information is required!"
                    quit()

            # automatic choise of temperature profile
            doy = time_slot.timetuple().tm_yday
            print "... automatic choise of temperature profile lon=", lon, " lat=", lat, ", time=", str(time_slot), ", doy=", doy
            if abs(lat) <= 30.0:
                cth_atm = "tropics"
            elif doy < 80 or doy <= 264:
                # northern summer
                if lat < -60.0:
                    cth_atm = "subarctic winter"
                elif -60.0 <= lat and lat < -30.0:
                    cth_atm = "midlatitude winter"
                elif 30.0 < lat and lat <= 60.0:
                    cth_atm = "midlatitude summer"
                elif 60.0 < lat:
                    cth_atm = "subarctic summer"
            else:
                # northern winter
                if lat < -60.0:
                    cth_atm = "subarctic summer"
                elif -60.0 <= lat and lat < -30.0:
                    cth_atm = "midlatitude summer"
                elif 30.0 < lat and lat <= 60.0:
                    cth_atm = "midlatitude winter"
                elif 60 < lat:
                    cth_atm = "subarctic winter"
            print "    choosing temperature profile for ", cth_atm

        # estimate cloud top height by searching first fit of ir108 with
        # temperature profile
        from mpop.tools import estimate_cth
        cth = estimate_cth(ir108, cth_atm=cth_atm)

        # create new channel named "CTH"
        self.channels.append(Channel(name="CTH",
                                     wavelength_range=[0., 0., 0.],
                                     resolution=self["IR_108"].resolution,
                                     data=cth,
                                     calibration_unit="m"))

        # copy additional information from IR_108
        self["CTH"].info = self["IR_108"].info
        self["CTH"].info['units'] = 'm'
        self["CTH"].area = self["IR_108"].area
        self["CTH"].area_id = self["IR_108"].area_id
        self["CTH"].area_def = self["IR_108"].area_def
        self["CTH"].resolution = self["IR_108"].resolution

        return cth

    def parallax_corr(self, fill="False", estimate_cth=False, cth_atm='best', replace=False):
        """
        perform the CTH parallax corretion for all loaded channels
        """

        loaded_channels = [chn.name for chn in self.loaded_channels()]
        if len(loaded_channels) == 0:
            return

        # loop over channels and check, if one is a normal radiance channel
        # having the method to calculate the viewing geometry
        for chn in self.loaded_channels():
            if hasattr(chn, 'get_viewing_geometry'):
                # calculate the viewing geometry of the SEVIRI sensor
                print "... calculate viewing geometry using ", chn.name
                (azi, ele) = chn.get_viewing_geometry(
                    self.get_orbital(), self.time_slot)
                break

        # choose best way to get CTH for parallax correction
        if not estimate_cth:
            if "CTTH" in loaded_channels:
                # make a copy of CTH, as it might get replace by its parallax
                # corrected version
                cth = copy.deepcopy(self["CTTH"].height)
            else:
                print "*** Error in parallax_corr (mpop.scene.py)"
                print "    parallax correction needs some cloud top height information"
                print "    please load the NWC-SAF CTTH product (recommended) or"
                print "    activate the option data.parallax_corr( estimate_cth=True )"
                quit()
        else:
            if "IR_108" in loaded_channels:
                # try to estimate CTH with IR_108
                self.estimate_cth()
                cth = self["CTH"].data
            else:
                print "*** Error in parallax_corr (mpop.scene.py)"
                print "    parallax correction needs some cloud top height information"
                print "    you specified the estimation of CTH with the IR_108, but "
                print "    this channel is not loaded"
                quit()

        # perform parallax correction for each loaded channel
        for chn in self.loaded_channels():
            if hasattr(chn, 'parallax_corr'):
                print "... perform parallax correction for ", chn.name
                if replace:
                    chn_name_PC = chn.name
                    print "    replace channel ", chn_name_PC
                else:
                    chn_name_PC = chn.name + "_PC"
                    print "    create channel ", chn_name_PC

                # take care of the parallax correction
                self[chn_name_PC] = chn.parallax_corr(
                    cth=cth, azi=azi, ele=ele, fill=fill)
            else:
                LOG.warning("Channel " + str(chn.name) + " has no attribute parallax_corr,"
                            "thus parallax effect wont be corrected.")
                print "Channel " + str(chn.name) + " has no attribute parallax_corr,"
                print "thus parallax effect wont be corrected."

        return self

    def project(self, dest_area, channels=None, precompute=False, mode=None,
                radius=None, nprocs=1):
        """Make a copy of the current snapshot projected onto the
        *dest_area*. Available areas are defined in the region configuration
        file (ACPG). *channels* tells which channels are to be projected, and
        if None, all channels are projected and copied over to the return
        snapshot.

        If *precompute* is set to true, the projecting data is saved on disk
        for reusage. *mode* sets the mode to project in: 'quick' which works
        between cartographic projections, and, as its denomination indicates,
        is quick (but lower quality), and 'nearest' which uses nearest
        neighbour for best projection. A *mode* set to None uses 'quick' when
        possible, 'nearest' otherwise.

        *radius* defines the radius of influence for neighbour search in
        'nearest' mode (in metres). Setting it to None, or omitting it will
        fallback to default values (5 times the channel resolution) or 10,000m
        if the resolution is not available.

        Note: channels have to be loaded to be projected, otherwise an
        exception is raised.
        """

        if not is_pyresample_loaded:
            # Not much point in proceeding then
            return self

        _channels = set([])

        if channels is None:
            for chn in self.loaded_channels():
                _channels |= set([chn])

        elif isinstance(channels, (list, tuple, set)):
            for chn in channels:
                try:
                    _channels |= set([self[chn]])
                except KeyError:
                    LOG.warning("Channel " + str(chn) + " not found,"
                                "thus not projected.")
        else:
            raise TypeError("Channels must be a list/"
                            "tuple/set of channel keys!")

        res = copy.copy(self)

        if isinstance(dest_area, str):
            dest_area = mpop.projector.get_area_def(dest_area)

        res.area = dest_area
        res.channels = []

        if not _channels <= self.loaded_channels():
            LOG.warning("Cannot project nonloaded channels: %s.",
                        _channels - self.loaded_channels())
            LOG.info("Will project the other channels though.")
            _channels = _channels and self.loaded_channels()

        cov = {}

        for chn in sorted(_channels, key=lambda x: x.resolution, reverse=True):
            if chn.area is None:
                if self.area is None:
                    area_name = ("swath_" + self.fullname + "_" +
                                 str(self.time_slot) + "_"
                                 + str(chn.shape))
                    chn.area = area_name
                else:
                    if is_pyresample_loaded:
                        try:
                            chn.area = AreaDefinition(
                                self.area.area_id + str(chn.shape),
                                self.area.name,
                                self.area.proj_id,
                                self.area.proj_dict,
                                chn.shape[1],
                                chn.shape[0],
                                self.area.area_extent,
                                self.area.nprocs)

                        except AttributeError:
                            try:
                                dummy = self.area.lons
                                dummy = self.area.lats
                                chn.area = self.area
                                area_name = ("swath_" + self.fullname + "_" +
                                             str(self.time_slot) + "_"
                                             + str(chn.shape))
                                chn.area.area_id = area_name
                            except AttributeError:
                                chn.area = self.area + str(chn.shape)
                    else:
                        chn.area = self.area + str(chn.shape)
            else:  # chn.area is not None
                # if (is_pyresample_loaded and
                #     (not hasattr(chn.area, "area_id") or
                #      not chn.area.area_id)):
                #     area_name = ("swath_" + self.fullname + "_" +
                #                  str(self.time_slot) + "_"
                #                  + str(chn.shape) + "_"
                #                  + str(chn.name))
                #     chn.area.area_id = area_name

                # This leaks memory !
                #LOG.debug("chn.area = " + str(chn.area))
                #LOG.debug("type(chn.area) = " + str(type(chn.area)))
                if is_pyresample_loaded:
                    area_name = ("swath_" + self.fullname + "_" +
                                 str(self.time_slot) + "_" +
                                 str(chn.shape) + "_" +
                                 str(chn.name))
                    LOG.debug("pyresample is loaded... area-name = " +
                              str(area_name))
                    if hasattr(chn.area, "area_id") and not chn.area.area_id:
                        LOG.debug("chn.area has area_id attribute...")
                        chn.area.area_id = area_name
                    elif not hasattr(chn.area, "area_id") and not isinstance(chn.area, str):
                        setattr(chn.area, 'area_id', area_name)

            if isinstance(chn.area, str):
                area_id = chn.area
            else:
                area_id = chn.area_id or chn.area.area_id

            if area_id not in cov:
                if radius is None:
                    if chn.resolution > 0:
                        radius = 5 * chn.resolution
                    else:
                        radius = 10000
                cov[area_id] = mpop.projector.Projector(chn.area,
                                                        dest_area,
                                                        mode=mode,
                                                        radius=radius,
                                                        nprocs=nprocs)
                if precompute:
                    try:
                        cov[area_id].save()
                    except IOError:
                        LOG.exception("Could not save projection.")

            try:
                res.channels.append(chn.project(cov[area_id]))
            except NotLoadedError:
                LOG.warning("Channel " + str(chn.name) + " not loaded, "
                            "thus not projected.")

        # Compose with image object
        try:
            if res._CompositerClass is not None:
                # Pass weak ref to compositor to allow garbage collection
                res.image = res._CompositerClass(weakref.proxy(res))
        except AttributeError:
            pass

        return res

if sys.version_info < (2, 5):
    def any(iterable):
        for element in iterable:
            if element:
                return True
        return False


def assemble_segments(segments):
    """Assemble the scene objects listed in *segment_list* and returns the
    resulting scene object.
    """
    from mpop.satellites import GenericFactory

    channels = set([])
    for seg in segments:
        channels |= set([chn.name for chn in seg.loaded_channels()])

    seg = segments[0]

    new_scene = GenericFactory.create_scene(seg.satname, seg.number,
                                            seg.instrument_name, seg.time_slot,
                                            seg.orbit, variant=seg.variant)

    swath_definitions = {}

    for chn in channels:
        new_scene[chn] = np.ma.concatenate([seg[chn].data
                                            for seg in segments
                                            if seg[chn].is_loaded()])
        try:

            area_names = tuple([seg[chn].area.area_id
                                for seg in segments
                                if seg[chn].is_loaded()])
            if area_names not in swath_definitions:

                lons = np.ma.concatenate([seg[chn].area.lons[:]
                                          for seg in segments
                                          if seg[chn].is_loaded()])
                lats = np.ma.concatenate([seg[chn].area.lats[:]
                                          for seg in segments
                                          if seg[chn].is_loaded()])
                new_scene[chn].area = SwathDefinition(lons=lons, lats=lats)
                area_name = "+".join(area_names)
                new_scene[chn].area.area_id = area_name
                new_scene[chn].area_id = area_name
                swath_definitions[area_names] = new_scene[chn].area
            else:
                new_scene[chn].area = swath_definitions[area_names]
                new_scene[chn].area_id = new_scene[chn].area.area_id
        except AttributeError:
            pass

    try:
        lons = np.ma.concatenate([seg.area.lons[:] for seg in segments])
        lats = np.ma.concatenate([seg.area.lats[:] for seg in segments])
        new_scene.area = SwathDefinition(lons=lons, lats=lats)
        for chn in channels:
            if any([seg[chn].area for seg in segments]):
                try:
                    lon_arrays = []
                    lat_arrays = []
                    for seg in segments:
                        if seg[chn].area is not None:
                            lon_arrays.append(seg[chn].area.lons[:])
                            lat_arrays.append(seg[chn].area.lats[:])
                        else:
                            lon_arrays.append(seg.area.lons[:])
                            lat_arrays.append(seg.area.lats[:])
                    lons = np.ma.concatenate(lon_arrays)
                    lats = np.ma.concatenate(lat_arrays)
                    new_scene[chn].area = SwathDefinition(lons=lons, lats=lats)
                except AttributeError:
                    pass
    except AttributeError:
        pass

    return new_scene



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011, 2014, 2015, 2016.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""mpop.satellites is the module englobes all satellite specific modules. In
itself, it hold the mighty :meth:`mpop.satellites.get_satellite_class` method.
"""
import os.path
import weakref
from ConfigParser import ConfigParser, NoSectionError, NoOptionError
import logging

from mpop import CONFIG_PATH
from mpop.scene import SatelliteInstrumentScene


LOG = logging.getLogger(__name__)


def get_custom_composites(name):
    """Get the home made methods for building composites for a given satellite
    or instrument *name*.
    """
    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, "mpop.cfg"))
    try:
        module_name = conf.get("composites", "module")
    except (NoSectionError, NoOptionError):
        LOG.debug("No custom composites provided i config file...")
        return []

    LOG.debug("module_name = %s", str(module_name))
    try:
        name = name.replace("/", "")
        module = __import__(module_name, globals(), locals(), [name])
    except ImportError:
        LOG.debug("Failed to import custom compositer for %s", str(name))
        return []

    try:
        return getattr(module, name)
    except AttributeError:
        LOG.debug("Could not get attribute %s from %s", str(name), str(module))
        return []


def get_sat_instr_compositer((satellite, number, variant), instrument):
    """Get the compositer class for a given satellite, defined by the three
    strings *satellite*, *number*, and *variant*, and *instrument*. The class
    is then filled with custom composites if there are any (see
    :func:`get_custom_composites`). If no class is found, an attempt is made to
    build the class from a corresponding configuration file, see
    :func:`build_sat_instr_compositer`.
    """

    module_name = variant + satellite + number
    class_name = (variant.capitalize() + satellite.capitalize() +
                  number.capitalize() + instrument.capitalize())

    try:
        module = __import__(module_name, globals(), locals(), [class_name])
        klass = getattr(module, class_name)
        for k in get_custom_composites(instrument):
            klass.add_method(k)
        return klass
    except (ImportError, AttributeError):
        return build_sat_instr_compositer((satellite, number,
                                           variant),
                                          instrument)


def build_instrument_compositer(instrument_name):
    """Automatically generate an instrument compositer class from its
    *instrument_name*. The class is then filled with custom composites if there
    are any (see :func:`get_custom_composites`)
    """

    from mpop.instruments.visir import VisirCompositer
    instrument_class = type(instrument_name.capitalize() + "Compositer",
                            (VisirCompositer,),
                            {"instrument_name": instrument_name})
    for i in get_custom_composites(instrument_name):
        instrument_class.add_method(i)

    return instrument_class


def build_sat_instr_compositer((satellite, number, variant), instrument):
    """Build a compositer class for the given satellite (defined by the three
    strings *satellite*, *number*, and *variant*) and *instrument* on the fly,
    using data from a corresponding config file. They inherit from the
    corresponding instrument class, which is also created on the fly is no
    predefined module (containing a compositer) for this instrument is
    available (see :func:`build_instrument_compositer`).
    """

    fullname = variant + satellite + number

    conf = ConfigParser()
    config_file = os.path.join(CONFIG_PATH, fullname + ".cfg")
    LOG.debug("Looking for config file %s", config_file)
    if not os.path.exists(config_file):
        LOG.error("Can't find config file %s, is PPP_CONFIG_DIR set?",
                  config_file)
    conf.read(config_file)

    try:
        LOG.debug('Instrument: ' + str(instrument))
        mod = __import__("mpop.instruments." + instrument,
                         globals(), locals(),
                         [instrument.capitalize() + 'Compositer'])
        instrument_class = getattr(mod, instrument.capitalize() + 'Compositer')
        for i in get_custom_composites(instrument):
            # LOG.debug("Custom composite = " + str(i))
            instrument_class.add_method(i)

    except (ImportError, AttributeError):
        LOG.debug("Build instrument compositer: " + str(instrument))
        instrument_class = build_instrument_compositer(instrument)

    sat_class = type(str(variant.capitalize() +
                         satellite.capitalize() +
                         number.capitalize() +
                         instrument.capitalize() +
                         "Compositer"),
                     (instrument_class,),
                     {})

    for i in get_custom_composites(instrument):
        LOG.debug("Add method to sat_class: " + str(i))
        sat_class.add_method(i)

    return sat_class


class GeostationaryFactory(object):

    """Factory for geostationary satellite scenes.
    """

    @staticmethod
    def create_scene(satname, satnumber, instrument, time_slot, area=None,
                     variant=''):
        """Create a compound satellite scene.
        """

        return GenericFactory.create_scene(satname, satnumber, instrument,
                                           time_slot, None, area, variant)


class PolarFactory(object):

    """Factory for polar satellite scenes.
    """

    @staticmethod
    def create_scene(satname, satnumber, instrument, time_slot, orbit=None,
                     area=None, variant=''):
        """Create a compound satellite scene.
        """

        return GenericFactory.create_scene(satname, satnumber, instrument,
                                           time_slot, orbit, area, variant)


class GenericFactory(object):

    """Factory for generic satellite scenes.
    """

    @staticmethod
    def create_scene(satname, satnumber, instrument, time_slot, orbit,
                     area=None, variant=''):
        """Create a compound satellite scene.
        """

        satellite = (satname, satnumber, variant)

        instrument_scene = SatelliteInstrumentScene(satellite=satellite,
                                                    instrument=instrument,
                                                    area=area,
                                                    orbit=orbit,
                                                    time_slot=time_slot)

        compositer = get_sat_instr_compositer(satellite, instrument)
        instrument_scene._CompositerClass = compositer

        if compositer is not None:
            # Pass weak ref to compositor to allow garbage collection
            instrument_scene.image = compositer(
                weakref.proxy(instrument_scene))
        return instrument_scene



import matplotlib as mpl   # this HAS TO BE the very first lines (before any other matplotlib functions are imported) 
mpl.use('Agg')             # this HAS TO BE the very first lines (before any other matplotlib functions are imported) 
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from matplotlib import rcParams
from PIL import Image as PIL_Image
from TRTimage import fig2data, fig2img
from pylab import text as pylab_text
from numpy import sin, cos, radians, where, nonzero, transpose, arange, append, meshgrid, mgrid, empty, isnan, nan, percentile
from numpy import sum as np_sum
from scipy.interpolate import griddata  # interp2d
from matplotlib.patches import Rectangle
from matplotlib.colors import Normalize

def prepare_figure(obj_area):

    # create new figure 
    #fig = Figure()  # old version, does not work for the stream plot 

    ## Turn interactive plotting off
    #plt.ioff()
    fig = plt.figure()   # needs a DISPLAY environment variable (simulated here with mpl.use('Agg'))

    # define size of image 
    nx = obj_area.x_size
    ny = obj_area.y_size
    # canvas figure 
    canvas = FigureCanvas(fig)
    # get dots per inch of the screen
    DPI = fig.get_dpi()
    # print "DPI", DPI
    fig.set_size_inches(nx/float(DPI),ny/float(DPI))
    # set fonts to bold
    plt.rc('font', weight='bold')
    # get axis object 
    ax = fig.add_subplot(111, aspect='equal')
    ## eliminates margins totally 
    fig.subplots_adjust(left=0.0,right=1.0,bottom=0.0,top=1.0, wspace=0, hspace=0)
    # set limits of the axis
    ax.set_xlim(0, nx)
    ax.set_ylim(0, ny)
    # set transparent backgroud 
    fig.patch.set_alpha(0.0)        # transparent outside of diagram  
    ax.set_axis_bgcolor([1,0,0,0])  # transparent color inside diagram 

    return fig, ax 


def HRWimage( HRW_data, obj_area, hrw_channels=None, min_correlation=None, cloud_type=None, style='barbs', \
                        barb_length=None, color_mode='channel', pivot='middle', legend=True, legend_loc=3):

    """Create a PIL image from high resolution wind data
       required input:
         HRW data [HRW_class]:      HRW_class instant containing the data, see mpop/satin/nwcsaf_hrw_hdf.py
         obj_area [area_class]:     instant of area class, returned by area_def
       optional input
       hrw_channels [string array]: giving the channels that are used derive the HRW vectors
                                    e.g. hrw_channels['HRV','WV_073']
       min_correlation [int]:       minimum correlation of tracking, if below arrow is not shown
       cloud_type [int array]:      cloud types of the wind vectors, e.g. cloud_type=[8,10,11]
       style [string]:              different styles of plotting
                                    style='barbs' or style='5min_displacement' or style='15min_displacement'
       color_mode [string]:         choose color of the wind symbols, possible choises:
                                    color_mode='channel'      -> one color per SEVIRI channel used to derive HRW
                                    color_mode='pressure'     -> colorcoded cloud top pressure
                                    color_mode='temperature'  -> colorcoded cloud top temperature
                                    color_mode='cloud_type'   -> NWC-SAF cloud types
                                    color_mode='correlation'  80 ... 100
                                    color_mode='conf_nwp'      70 ... 100
                                    color_mode='conf_no_nwp'   70 ... 100
       pivot [string]:              position of the barb, e.g. pivot='middle' == center of barb at origin
       legend [True or False] :     show legend or not 
       legend_loc [string or int]:  location of the legend 
                                    upper right    1
                                    upper left     2
                                    lower left     3
                                    lower right    4
                                    right          5
                                    center left    6
                                    center right   7
                                    lower center   8
                                    upper center   9
                                    center         10
                                    best
    """

    #if min_correlation != None:
    #    print "    filter for min_correlation = ", min_correlation
    #    inds = where(HRW_data.correlation > min_correlation)
    #    HRW_data = HRW_data.subset(inds)
     
    print "... create HRWimage, color_mode = ", color_mode

    # get a empty figure with transparent background, no axis and no margins outside the diagram
    fig, ax = prepare_figure(obj_area)

    # define arrow properties 
    head_width  = 0.006 * min(obj_area.x_size,obj_area.x_size)
    head_length = 2 * head_width
    m_per_s_to_knots = 1.944

    #barb_length = 0.008 * min(obj_area.x_size,obj_area.x_size)
    
    if barb_length == None:
        n_winds = len(HRW_data.wind_id)
        if n_winds < 300:
            barb_length = 5.68
        elif n_winds < 500:
            barb_length = 5.43
        elif n_winds < 700:
            barb_length = 5.18
        elif n_winds < 900:
            barb_length = 4.68
        else:          
            barb_length = 4.00
    print "barb_length", barb_length

    if color_mode == 'channel':
        classes = ('HRV',          'VIS008 ', 'WV_062 ',   'WV_073 ',   'IR_120 ')
        colors   = ['mediumorchid', 'red',     'limegreen', 'darkgreen', 'darkturquoise']
    elif color_mode == 'pressure':
        classes = ['<200hPa',  '200-300hPa','300-400hPa','400-500hPa','500-600hPa','600-700hPa','700-800hPa', '800-900hPa','>900hPa']
        colors   = ['darksalmon', 'red'     ,'darkorange','yellow'    ,'lime'      ,'seagreen',  'deepskyblue','blue',      'mediumorchid']
        classes = tuple(['CTP '+cl for cl in classes])
    elif color_mode == 'cloud_type' or color_mode == 'cloudtype':
        classes=['non-processed','cloud free land', 'cloud free sea', 'land snow', 'sea ice',\
                 'very low cum.', 'very low', 'low cum.', 'low', 'med cum.', 'med', 'high cum.', 'high', 'very high cum.', 'very high', \
                 'sem. thin', 'sem. med.', 'sem. thick', 'sem. above', 'broken', 'undefined']

        colors = empty( (len(classes),3), dtype=int )
        colors[ 0,:] = [100, 100, 100]
        colors[ 1,:] = [  0, 120,   0]
        colors[ 2,:] = [  0,   0,   0]
        colors[ 3,:] = [250, 190, 250]
        colors[ 4,:] = [220, 160, 220]
        colors[ 5,:] = [255, 150,   0]
        colors[ 6,:] = [255, 100,   0]
        colors[ 7,:] = [255, 220,   0]
        colors[ 8,:] = [255, 180,   0]
        colors[ 9,:] = [255, 255, 140]
        colors[10,:] = [240, 240,   0]
        colors[11,:] = [250, 240, 200]
        colors[12,:] = [215, 215, 150]
        colors[13,:] = [255, 255, 255]
        colors[14,:] = [230, 230, 230]
        colors[15,:] = [  0,  80, 215]
        colors[16,:] = [  0, 180, 230]
        colors[17,:] = [  0, 240, 240]
        colors[18,:] = [ 90, 200, 160]
        colors[19,:] = [200,   0, 200]
        colors[20,:] = [ 95,  60,  30]
        colors = colors/255.
    elif color_mode in ['correlation','conf_nwp','conf_no_nwp']:
        classes  = ['<70',    '<75',     '<80',    '<85',   '<90',  '<95',   '>95' ]
        colors   = ['indigo', 'darkred', 'red','darkorange','gold', 'lime', 'green']
        classes = tuple([color_mode+' '+cl for cl in classes])
    else:
          print "*** Error in HRW_streamplot (mpop/imageo/HRWimage.py)"
          print "    unknown color_mode"
          quit()

    for wid in range(len(HRW_data.wind_id)):

        if color_mode == 'channel':

            if HRW_data.channel[wid].find('HRV') != -1: # HRV
                barbcolor = colors[0]
            elif HRW_data.channel[wid].find('VIS008') != -1: #  0.8 micro m
                barbcolor = colors[1]
            elif HRW_data.channel[wid].find('WV_062') != -1: #  6.2 micro m
                barbcolor = colors[2]
            elif HRW_data.channel[wid].find('WV_073') != -1: #  7.3 micro m
                barbcolor = colors[3]
            elif HRW_data.channel[wid].find('IR_120') != -1: # 12.0 micro m
                barbcolor = colors[4]

        elif color_mode == 'pressure':

            if HRW_data.pressure[wid] < 20000:
                barbcolor = colors[0]
            elif HRW_data.pressure[wid] < 30000:
                barbcolor = colors[1]
            elif HRW_data.pressure[wid] < 40000:
                barbcolor = colors[2]
            elif HRW_data.pressure[wid] < 50000:
                barbcolor = colors[3] 
            elif HRW_data.pressure[wid] < 60000:
                barbcolor = colors[4]
            elif HRW_data.pressure[wid] < 70000:
                barbcolor = colors[5]
            elif HRW_data.pressure[wid] < 80000:
                barbcolor = colors[6]
            elif HRW_data.pressure[wid] < 90000:
                barbcolor = colors[7]
            else:
                barbcolor = colors[8]

        elif color_mode == 'cloud_type' or color_mode == 'cloudtype':

            barbcolor = list(colors[HRW_data.cloud_type[wid], :])

        elif color_mode in ['correlation','conf_nwp','conf_no_nwp']:
            if color_mode == 'correlation':
                cdata = HRW_data.correlation
            elif color_mode == 'conf_nwp':
                cdata = HRW_data.conf_nwp
            elif color_mode == 'conf_no_nwp':
                cdata = HRW_data.conf_no_nwp

            if cdata[wid] < 70:
                barbcolor = colors[0]
            elif cdata[wid] < 75:
                barbcolor = colors[1]
            elif cdata[wid] < 80:
                barbcolor = colors[2]
            elif cdata[wid] < 85:
                barbcolor = colors[3] 
            elif cdata[wid] < 90:
                barbcolor = colors[4]
            elif cdata[wid] < 95:
                barbcolor = colors[5]
            else:
                barbcolor = colors[6]
        else:
              print "*** Error in HRW_streamplot (mpop/imageo/HRWimage.py)"
              print "    unknown color_mode"
              quit()
      
        x0, y0 = obj_area.get_xy_from_lonlat( HRW_data.lon[wid], HRW_data.lat[wid], outside_error=False) #, return_int=True

        u = HRW_data.wind_speed[wid] * -1 * sin(radians(HRW_data.wind_direction[wid])) 
        v = HRW_data.wind_speed[wid] * -1 * cos(radians(HRW_data.wind_direction[wid]))

        #print '%6s %3d %10.7f %10.7f %7.2f %7.1f %8.1f %10s' % (HRW_data.channel[wid], HRW_data.wind_id[wid], \
        #                                                        HRW_data.lon[wid], HRW_data.lat[wid], \
        #                                                        HRW_data.wind_speed[wid]*m_per_s_to_knots, \
        #                                                        HRW_data.wind_direction[wid], HRW_data.pressure[wid], barbcolor)


        if style == 'barbs':
            u = HRW_data.wind_speed[wid] * -1 * sin(radians(HRW_data.wind_direction[wid])) * m_per_s_to_knots
            v = HRW_data.wind_speed[wid] * -1 * cos(radians(HRW_data.wind_direction[wid])) * m_per_s_to_knots
            ax.barbs(x0, obj_area.y_size - y0, u * m_per_s_to_knots, v * m_per_s_to_knots, length = barb_length, pivot='middle', barbcolor=barbcolor)

        elif style == '5min_displacement' or style == '15min_displacement':
            if style == '5min_displacement':
                t_in_s =  5*60
            elif style == '15min_displacement':
                t_in_s = 15*60
            dx = u * t_in_s / obj_area.pixel_size_x
            dy = v * t_in_s / obj_area.pixel_size_y
            ax.arrow(x0, y0, dx, dy, head_width = head_width, head_length = head_length, fc=barbcolor, ec=barbcolor)

    if legend:

        rcParams['legend.handlelength'] = 0
        rcParams['legend.numpoints'] = 1

        # create blank rectangle
        rec = Rectangle((0, 0), 0, 0, fc="w", fill=False, edgecolor='none', linewidth=0)

        ##  *fontsize*: [size in points | 'xx-small' | 'x-small' | 'small' |
        ##              'medium' | 'large' | 'x-large' | 'xx-large']


        alpha=1.0
        bbox={'facecolor':'white', 'alpha':alpha, 'pad':10}

        print "... add legend: color is a function of ",  color_mode
        
        recs = empty( len(classes), dtype=object)
        recs[:] = rec 

        #if color_mode == 'pressure':
        #    recs = [rec, rec, rec, rec, rec, rec, rec, rec, rec]
        #if color_mode == 'channel':
        #    recs = [rec, rec, rec, rec, rec]
        #if color_mode in ['correlation','conf_nwp','conf_no_nwp']:
        #    recs = [rec, rec, rec, rec, rec, rec, rec]

        size=12
        if color_mode=='cloud_type':
            size=10

        leg = ax.legend(recs, classes, loc=legend_loc, prop={'size':size})

        for color,text in zip(colors,leg.get_texts()):
            text.set_color(color)


    return fig2img ( fig )


# ----------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------

def HRWstreamplot( u2d, v2d, obj_area, interpol_method, color_mode='speed', density=(3,3), linewidth_mode="scaled", linewidth_max=2.5, \
                   min_correlation=None, legend=True, legend_loc=3, vmax=None, colorbar=True, fontcolor='w'):

    """ Create a streamplot image in PIL format of a 2d wind field
           color_mode [string]:     choose color of the stream lines,
                                    color_mode='speed' -> wind speed
                                    color_mode='u'     -> u-wind component
                                    color_mode='v'     -> v-wind component
           density [2 int tuple]    density of stream lines, default density = (4,4)
           linewidth_mode [string]  "scaled" to color_mode data
                                    "const" always linewith_max
    """


    ## get a empty figure with transparent background, no axis and no margins outside the diagram
    fig, ax = prepare_figure(obj_area)
    #print dir(ax)

    # check if there is there have been enough observation (or if data is not a number array)
    if isnan(np_sum(u2d)):
        print "... there are not enough observations"
        ax.text(0.95, 0.01, 'currently not enough observations',
                verticalalignment='bottom', horizontalalignment='right',
                transform=ax.transAxes, color='red', fontsize=15)
    else:
        print "there is enough data, interpolation method: ", interpol_method

        # create grid for the wind data
        [nx, ny] = u2d.shape  # for ccs4 this is (640, 710)
        Y, X = mgrid[nx-1:-1:-1, 0:ny] # watch out for Y->nx and X->ny
        #print "X.shape ", Y.shape
        #print Y[:,0]

        print "   calculate color data ", color_mode
        if color_mode == 'speed':
            from numpy import sqrt
            cdata = sqrt(u2d*u2d + v2d*v2d)
        elif color_mode == 'u':
            cdata = u2d
        elif color_mode == 'v':
            cdata = v2d
        else:
            print "*** Error in HRW_streamplot (mpop/imageo/HRWimage.py)"
            print "    unknown color_mode"
            quit()

        print "   calculate linewidth ", linewidth_mode
        if linewidth_mode == "const":
            linewidth = linewidth_max
        elif linewidth_mode == "scaled":
            if vmax != None:
                linewidth = 1 + linewidth_max*(cdata) / vmax
            else:
                linewidth = 1 + linewidth_max*(cdata) / cdata.max()
        else:
            print "*** Error in HRW_streamplot (mpop/imageo/HRWimage.py)"
            print "    unknown linewidth_mode"
            quit()

        print "... data_max =", cdata.max() ,", vmax=", vmax

        if vmax != None:
            norm = Normalize(vmin=0, vmax=vmax)
        else:
            norm = Normalize(vmin=0, vmax=cdata.max())

        #optional arguments of streamplot
        #           density=1, linewidth=None, color=None,
        #           cmap=None, norm=None, arrowsize=1, arrowstyle='-|>',
        #           minlength=0.1, transform=None, zorder=1, start_points=None, INTEGRATOR='RK4',density=(10,10)
        plt.streamplot(X, Y, u2d, v2d, color=cdata, linewidth=linewidth, cmap=plt.cm.rainbow, density=density, norm=norm) 

        if colorbar:
            colorbar_ax = fig.add_axes([0.9, 0.1, 0.05, 0.8])
            cbar = plt.colorbar(cax=colorbar_ax)
            plt.setp(plt.getp(cbar.ax.axes, 'yticklabels'), color=fontcolor)
            xlabel = cbar.ax.set_xlabel('m/s', weight='bold') #get the title property handler
            plt.setp(xlabel, color=fontcolor) 

        #plt.savefig("test_streamplot.png")

        # add information about interpolation method
        ax.text(0.95, 0.01, interpol_method,
                verticalalignment='bottom', horizontalalignment='right',
                transform=ax.transAxes, color='green', fontsize=15)

    return fig2img ( fig )


# ----------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------


def fill_with_closest_pixel(data, invalid=None):
    """
    Replace the value of invalid 'data' cells (indicated by 'invalid') 
    by the value of the nearest valid data cell

    Input:
        data:    numpy array of any dimension
        invalid: a binary array of same shape as 'data'. True cells set where data
                 value should be replaced.
                 If None (default), use: invalid  = np.isnan(data)

    Output: 
        Return a filled array. 
    """

    from numpy import isnan 
    from scipy.ndimage import distance_transform_edt

    if invalid is None: invalid = isnan(data)

    ind = distance_transform_edt(invalid, return_distances=False, return_indices=True)

    return data[tuple(ind)]

# ----------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------
def HRWscatterplot( HRW_data, title='', hrw_channels=None, min_correlation=None, cloud_type=None, color_mode='direction'):

    ## get a empty figure with transparent background, no axis and no margins outside the diagram
    # fig = plt.figure()
    import pylab 
    fig = pylab.figure()
    ax = plt.subplot(111)
    ax.set_yscale("log", nonposx='clip')
    plt.scatter(HRW_data.wind_speed, HRW_data.pressure/100, s=5, c=HRW_data.wind_direction, alpha=0.5, edgecolor='none')
    pylab.title(title)
    pylab.ylim([1000,100])
    plt.yticks([1000,900,800,700,600,500,400,300,200,100], ['1000','900','800','700','600','500','400','300','200','100'], rotation='horizontal')

    p = percentile(HRW_data.wind_speed, 95)
    vmax = (round(p/10)+1)*10
    print "... vmax:", vmax 

    plt.plot([0,vmax], [680,680], color='g')
    plt.plot([0,vmax], [440,440], color='b')

    pylab.xlim([0,vmax])
    ax.set_xlabel('HRW [m/s]')
    ax.set_ylabel('p [hPa]')

    cbar = plt.colorbar()
    cbar.ax.set_ylabel('wind direction')

    return fig2img ( fig )

# ----------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------

def HRW_2dfield( HRW_data, obj_area, interpol_method=None, hrw_channels=None, min_correlation=None, level=''):

    print "... calculate 2d wind field (HRW_2dfield)"

    if min_correlation != None:
        print "    filter for min_correlation = ", min_correlation
        inds = where(HRW_data.correlation > min_correlation)
        HRW_data.subset(inds)

    xx, yy = obj_area.get_xy_from_lonlat( HRW_data.lon, HRW_data.lat, outside_error=False, return_int=False) #, return_int=True

    yy = obj_area.y_size - yy  

    uu = - HRW_data.wind_speed * sin(radians(HRW_data.wind_direction))
    vv = - HRW_data.wind_speed * cos(radians(HRW_data.wind_direction))

    # get rid of all vectors outside of the field 
    index = nonzero(xx)
    xx = xx[index]
    yy = yy[index]
    uu = uu[index]
    vv = vv[index]

    points = transpose(append([xx], [yy], axis=0))
    #print type(uu), uu.shape
    #print type(points), points.shape
    #print points[0], yy[0], xx[0]
    #print uu[0]

    nx = obj_area.x_size
    ny = obj_area.y_size

    x2 = arange(nx)
    y2 = (ny-1) - arange(ny)

    grid_x, grid_y = meshgrid(x2, y2)
    
    if interpol_method == None:
        # we need at least 2 winds to interpolate 
        if uu.size < 4:  
            print "*** Warning, not wnough wind data available, n_winds = ", uu.size
            fake = empty(grid_x.shape)
            fake[:,:] = nan
            HRW_data.interpol_method = None
            return fake, fake

        elif uu.size < 50:
            interpol_method = "RBF"

        else:
            interpol_method = "linear + nearest"
     
    #interpol_method = "nearest"
    #interpol_method = "cubic + nearest" # might cause unrealistic overshoots
    #interpol_method = "kriging"
    #interpol_method = "..."

    print "... min windspeed (org data): ", HRW_data.wind_speed.min()
    print "... max windspeed (org data): ", HRW_data.wind_speed.max()

    for i_iteration in [0,1]:

        if interpol_method == "nearest":

            print '... fill with nearest neighbour'
            # griddata, see http://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.interpolate.griddata.html
            grid_u1x = griddata(points, uu, (grid_x, grid_y), method='nearest')
            grid_v1x = griddata(points, vv, (grid_x, grid_y), method='nearest')

        elif interpol_method == "RBF":

            print '... inter- and extrapolation using radial basis functions'
            # https://www.youtube.com/watch?v=_cJLVhdj0j4
            print "... start Rbf"
            from scipy.interpolate import Rbf
            # rbfu = Rbf(xx, yy, uu, epsilon=0.1) #
            rbfu = Rbf(xx, yy, uu, epsilon=0.2)
            grid_u1x = rbfu(grid_x, grid_y)
            rbfv = Rbf(xx, yy, vv, epsilon=0.1) #
            grid_v1x = rbfv(grid_x, grid_y)
            print "... finish Rbf"
            # !very! slow for a large number of observations 

        elif interpol_method == "linear + nearest" or interpol_method == "cubic + nearest":

            if interpol_method == "linear + nearest":
                print '... calculate linear interpolation'
                # griddata, see http://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.interpolate.griddata.html
                grid_u1 = griddata(points, uu, (grid_x, grid_y), method='linear')
                grid_v1 = griddata(points, vv, (grid_x, grid_y), method='linear')
            elif interpol_method == "cubic + nearest":
                # smoother, but can cause unrealistic overshoots
                print '... calculate cubic interpolation'
                grid_u1 = griddata(points, uu, (grid_x, grid_y), method='cubic')
                grid_v1 = griddata(points, vv, (grid_x, grid_y), method='cubic')
            else:
                print "*** Error in mpop/imageo/HRWimage.py"
                print "    unknown interpolation method: ", interpol_method
                quit()

            if 1==1:
                # use faster function to extrapolate with closest neighbour
                print "... fill outside area with closest value"
                grid_u1x = fill_with_closest_pixel(grid_u1, invalid=None) 
                grid_v1x = fill_with_closest_pixel(grid_v1, invalid=None) 
            else:
                # use griddata to extrapolate with closest neighbour
                points2 = transpose(append([grid_x.flatten()], [grid_y.flatten()], axis=0))
                print type(grid_x.flatten()), grid_x.flatten().shape
                print type(points2), points2.shape
                mask = ~isnan(grid_v1.flatten())
                inds = where(mask)[0]
                grid_u1x = griddata(points2[inds], grid_u1.flatten()[inds], (grid_x, grid_y), method='nearest')
                grid_v1x = griddata(points2[inds], grid_v1.flatten()[inds], (grid_x, grid_y), method='nearest')

            if 1==0:
                # add othermost points as additional data
                y_add = [0,    0, ny-1, ny-1]
                x_add = [0, nx-1,    0, nx-1]
                for (i,j) in zip(x_add,y_add):
                    uu = append(uu, grid_u0[i,j])
                    vv = append(vv, grid_v0[i,j])
                xx = append(xx, x_add)
                yy = append(yy, y_add)
                points = transpose(append([yy], [xx], axis=0))

                print 'calc extent1'
                grid_u1e = griddata(points, uu, (grid_x, grid_y), method='linear')
                grid_v1e = griddata(points, vv, (grid_x, grid_y), method='linear')

        else:
            print "*** Error in mpop/imageo/HRWimage.py"
            print "    unknown interpol_method", interpol_method
            quit()

        ##http://docs.scipy.org/doc/scipy/reference/tutorial/interpolate.html
        ##http://stackoverflow.com/questions/3526514/problem-with-2d-interpolation-in-scipy-non-rectangular-grid
        #print "SmoothBivariateSpline:"
        #from scipy.interpolate import SmoothBivariateSpline
        #fitu = SmoothBivariateSpline( xx, yy, uu, s=1000) # , kx=3, ky=3, s = smooth * z2sum m 
        #from numpy import empty 
        #grid_u_SBP = empty(grid_x.shape)
        #for i in range(0,nx-1):       # starting upper right going down
        #    for j in range(0,ny-1):   # starting lower right going right
        #        #print i,j
        #        grid_u_SBP[j,i] = fitu(j,i)

        #grid_u_SBP = np.array([k.predict([x,y]) for x,y in zip(np.ravel(grid_x), np.ravel(grid_y))])
        #grid_u_SBP = grid_u_SBP.reshape(grid_x.shape)

        ##print x2
        ##print y2
        #grid_u_SBP = fitu(x2,y2)
        ##print "grid_u_SBP.shape", grid_u_SBP.shape
        ###print grid_u_SBP
        #print "End SmoothBivariateSpline:"

        #print "bisplrep:"
        #from scipy import interpolate
        #tck = interpolate.bisplrep(xx, yy, uu)
        #grid_u_BSR = interpolate.bisplev(grid_x[:,0], grid_y[0,:], tck)
        #print grid_u_BSR.shape
        #print "bisplrep"
        #print "grid_v1x.shape", grid_v1x.shape

        extent=(0,nx,0,ny)
        origin='lower'
        origin='upper'
        origin=None 

        # show different stages of 2d inter- and extra-polation 
        if 1==0:
            print 'make matplotlib.pyplot'
            import matplotlib.pyplot as plt
            vmin=-10
            vmax=10
            fig = plt.figure()
            plt.subplot(221)
            plt.title('u '+interpol_method)
            plt.plot(points[:,0], ny-1-points[:,1], 'k.', ms=1)
            plt.imshow(grid_u1x, vmin=vmin, vmax=vmax) #, extent=extent
            #plt.colorbar()
            plt.subplot(222)
            plt.title('v '+interpol_method)
            plt.plot(points[:,0], ny-1-points[:,1], 'k.', ms=1)
            plt.imshow(grid_v1x, origin=origin, vmin=vmin, vmax=vmax) #, extent=extent
            #plt.colorbar()

            # standard calculation for comparison 
            print '... calculate linear interpolation'
            grid_u1 = griddata(points, uu, (grid_x, grid_y), method='linear')
            grid_v1 = griddata(points, vv, (grid_x, grid_y), method='linear')
            grid_u1xx = fill_with_closest_pixel(grid_u1, invalid=None) 
            grid_v1xx = fill_with_closest_pixel(grid_v1, invalid=None) 

            plt.subplot(223)
            plt.title('U Linear+Nearest')
            plt.plot(points[:,0], ny-1-points[:,1], 'k.', ms=1)
            plt.imshow(grid_u1xx, origin=origin, vmin=vmin, vmax=vmax) #, extent=extent
            #plt.colorbar()
            plt.subplot(224)
            plt.title('V Linear+Nearest') 
            plt.plot(points[:,0], ny-1-points[:,1], 'k.', ms=1)
            #plt.title('Cubic')
            plt.imshow(grid_v1xx, origin=origin, vmin=vmin, vmax=vmax) #, extent=extent
            #plt.colorbar()
            plt.gcf().set_size_inches(6, 6)
            #plt.show()  # does not work with AGG
            tmpfile="test_hrw"+level+".png"
            fig.savefig(tmpfile)
            print "display "+tmpfile+" &"


        if grid_u1x.min() < -150 or grid_v1x.min() < -150 or grid_u1x.max() > 150 or grid_v1x.max() > 150:
            print "*** Warning, numerical instability detected, interpolation method: ", interpol_method
            print "    min u windspeed (u 2dimensional): ", grid_u1x.min()
            print "    min v windspeed (v 2dimensional): ", grid_v1x.min()
            print "    max u windspeed (u 2dimensional): ", grid_u1x.max()
            print "    max v windspeed (v 2dimensional): ", grid_v1x.max()
            interpol_method = "glinear + nearest"
            print "... try another interpolation method: ", interpol_method
        else:
            # (hopefully) numerical stable interpolation, exit the interpolation loop
            break 

    HRW_data.interpol_method = interpol_method

    return grid_u1x, grid_v1x



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017
#
# Author(s):
#
#   Panu Lahtinen <panu.lahtinen@fmi.fi>
#
# This file is part of mpop.
#
# mpop is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# mpop is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Module for calculating and projection mapping look-up tables, which
can be used with MPOP by setting 'precompute=True'"""

import numpy as np
import sys

from mpop.projector import (calc_nearest_params,
                            calc_bilinear_params,
                            calc_quick_params,
                            get_precompute_cache_fname,
                            get_area_and_id)


def save(fname, **data):
    """Save preprojection data to npz file."""
    np.savez(fname, **data)


def calc_preproj_params(out_dir, mode, in_area_name, out_area_name,
                        radius=None, nprocs=1):
    """Calculate preprojection parameters and save to *out_dir*"""
    in_area, in_id = get_area_and_id(in_area_name)
    out_area, out_id = get_area_and_id(out_area_name)

    fname = get_precompute_cache_fname(in_id, out_id,
                                       in_area_name, out_area_name,
                                       mode, out_dir)

    if mode == "nearest":
        data = calc_nearest_params(in_area, out_area,
                                   radius, nprocs=nprocs)
    elif mode == "bilinear":
        data = calc_bilinear_params(in_area, out_area, radius, nprocs=nprocs)
    elif mode == "quick":
        data = calc_quick_params(in_area, out_area)

    save(fname, **data)


def print_usage():
    """Print usage"""
    print("USAGE:")
    print("python precompute_projection.py <in_area_name> "
          "<out_area_name>")
    print("python precompute_projection.py <in_area_name> "
          "<out_area_name> <mode>")
    print("python precompute_projection.py <in_area_name> "
          "<out_area_name> <mode> <search_radius>")
    print("python precompute_projection.py <in_area_name> "
          "<out_area_name> <mode> <search_radius> <nprocs>")
    print("python precompute_projection.py <in_area_name> "
          "<out_area_name> <mode> <search_radius> <nprocs> <out_dir>")


def main():
    try:
        in_area_name = sys.argv[1]
        out_area_name = sys.argv[2]
    except IndexError:
        print_usage()
        return

    try:
        mode = sys.argv[3]
    except IndexError:
        mode = "nearest"
    try:
        radius = int(sys.argv[4])
    except IndexError:
        radius = 50000
    try:
        nprocs = int(sys.argv[5])
    except IndexError:
        nprocs = 1

    try:
        out_dir = sys.argv[6]
    except IndexError:
        out_dir = '.'

    calc_preproj_params(out_dir, mode, in_area_name, out_area_name,
                        radius=radius, nprocs=nprocs)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011, 2012, 2014, 2015.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for scene.py.
"""

from mock import MagicMock, patch
import sys
sys.modules['pyresample.geometry'] = MagicMock()

import ConfigParser
import datetime
import random
import unittest

import numpy as np

import mpop.projector
from mpop.channel import NotLoadedError
from mpop.scene import SatelliteScene, SatelliteInstrumentScene


def random_string(length,
                  choices="abcdefghijklmnopqrstuvwxyz"
                  "ABCDEFGHIJKLMNOPQRSTUVWXYZ"):
    """Generates a random string with elements from *set* of the specified
    *length*.
    """
    return "".join([random.choice(choices)
                    for dummy_itr in range(length)])

EPSILON = 0.0001
DUMMY_STRING = "test_plugin"


def patch_configparser():
    """Patch to fake ConfigParser.
    """
    class FakeConfigParser:

        """Dummy ConfigParser class.
        """

        def __init__(self, *args, **kwargs):
            pass

        def read(self, *args, **kwargs):
            """Dummy read method
            """
            del args, kwargs
            self = self

        def get(self, *args, **kwargs):
            """Dummy get method
            """
            del args, kwargs
            self = self
            return DUMMY_STRING

        def sections(self):
            """Dummy sections method
            """
            # return ["satellite", "udlptou-4"]
            raise ConfigParser.NoSectionError("Dummy sections.")

    ConfigParser.OldConfigParser = ConfigParser.ConfigParser
    ConfigParser.ConfigParser = FakeConfigParser


def unpatch_configparser():
    """Unpatch fake ConfigParser.
    """
    ConfigParser.ConfigParser = ConfigParser.OldConfigParser
    delattr(ConfigParser, "OldConfigParser")


def patch_projector():
    """Patch to fake projector.
    """
    class FakeProjector:

        """Dummy Projector class.
        """

        def __init__(self, *args, **kwargs):
            del args, kwargs
            self.out_area = None

        def project_array(self, arg):
            """Dummy project_array method.
            """
            return arg

    def fake_get_area_def(area):
        return area

    mpop.projector.OldProjector = mpop.projector.Projector
    mpop.projector.Projector = FakeProjector
    mpop.projector.old_get_area_def = mpop.projector.get_area_def
    mpop.projector.get_area_def = fake_get_area_def


def unpatch_projector():
    """Unpatch fake projector
    """
    mpop.projector.Projector = mpop.projector.OldProjector
    delattr(mpop.projector, "OldProjector")
    mpop.projector.get_area_def = mpop.projector.old_get_area_def
    delattr(mpop.projector, "old_get_area_def")


class TestSatelliteScene(unittest.TestCase):

    """Class for testing the SatelliteScene class.
    """

    scene = None

    def test_init(self):
        """Creation of a satellite scene.
        """

        self.scene = SatelliteScene()
        self.assertEquals(self.scene.satname, "")
        self.assertEquals(self.scene.number, "")
        self.assert_(self.scene.time_slot is None)
        self.assert_(self.scene.orbit is None)
        self.assert_(self.scene.area is None)
        self.assert_(self.scene.lat is None)
        self.assert_(self.scene.lon is None)

        # time_slot

        time_slot = datetime.datetime.now()

        self.scene = SatelliteScene(time_slot=time_slot)
        self.assertEquals(self.scene.satname, "")
        self.assertEquals(self.scene.number, "")
        self.assertEquals(self.scene.time_slot, time_slot)
        self.assert_(self.scene.orbit is None)
        self.assert_(self.scene.area is None)
        self.assert_(self.scene.lat is None)
        self.assert_(self.scene.lon is None)

        # area

        area = random_string(int(np.random.uniform(9)) + 1)

        self.scene = SatelliteScene(area=area)
        self.assertEquals(self.scene.satname, "")
        self.assertEquals(self.scene.number, "")
        self.assertEquals(self.scene.area, area)
        self.assert_(self.scene.orbit is None)
        self.assert_(self.scene.time_slot is None)
        self.assert_(self.scene.lat is None)
        self.assert_(self.scene.lon is None)

        self.assertRaises(TypeError,
                          SatelliteScene,
                          area=np.random.uniform(1000))

        self.assertRaises(TypeError,
                          SatelliteScene,
                          area=int(np.random.uniform(1000)))

        self.assertRaises(TypeError,
                          SatelliteScene,
                          area=[])

        # orbit

        orbit = random_string(int(np.random.uniform(9)) + 1)

        self.scene = SatelliteScene(orbit=orbit)
        self.assertEquals(self.scene.satname, "")
        self.assertEquals(self.scene.number, "")
        self.assertEquals(self.scene.orbit, orbit)
        self.assert_(self.scene.area is None)
        self.assert_(self.scene.time_slot is None)
        self.assert_(self.scene.lat is None)
        self.assert_(self.scene.lon is None)

    def test_fullname(self):
        """Fullname of a sat scene.
        """
        self.scene = SatelliteScene()

        self.scene.satname = random_string(int(np.random.uniform(9)) + 1)
        self.scene.number = random_string(int(np.random.uniform(9)) + 1)
        self.scene.variant = random_string(int(np.random.uniform(9)) + 1)
        self.assertEquals(self.scene.fullname,
                          self.scene.variant +
                          self.scene.satname +
                          self.scene.number)


class TestSatelliteInstrumentScene(unittest.TestCase):

    """Class for testing the SatelliteInstrumentScene class.
    """

    scene = None

    def setUp(self):
        """Patch foreign modules.
        """
        patch_configparser()
        patch_projector()

    def test_init_area(self):
        """Creation of a satellite instrument scene.
        """
        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]

        class SatelliteInstrumentScene2(SatelliteInstrumentScene):

            """Dummy satinst class.
            """
            channel_list = channels
        # area

        area = random_string(int(np.random.uniform(9)) + 1)

        self.scene = SatelliteInstrumentScene2(area=area)
        self.assertEquals(self.scene.satname, "")
        self.assertEquals(self.scene.number, "")
        self.assertEquals(self.scene.area, area)
        self.assert_(self.scene.orbit is None)
        self.assert_(self.scene.time_slot is None)
        self.assert_(self.scene.lat is None)
        self.assert_(self.scene.lon is None)
        self.assert_(self.scene.instrument_name is None)
        self.assertEquals(self.scene.channels_to_load, set([]))
        for i, chn in enumerate(self.scene.channels):
            self.assertEquals(chn.name, channels[i][0])
            self.assertEquals(chn.wavelength_range, list(channels[i][1]))
            self.assertEquals(chn.resolution, channels[i][2])

        self.assertRaises(TypeError,
                          SatelliteInstrumentScene2,
                          area=np.random.uniform(1000))

        self.assertRaises(TypeError,
                          SatelliteInstrumentScene2,
                          area=int(np.random.uniform(1000)))

        self.assertRaises(TypeError,
                          SatelliteInstrumentScene2,
                          area=[])

    def test_init_orbit(self):
        """Creation of a satellite instrument scene.
        """
        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]

        class SatelliteInstrumentScene2(SatelliteInstrumentScene):

            """Dummy satinst class.
            """
            channel_list = channels

        # orbit

        orbit = random_string(int(np.random.uniform(9)) + 1)

        self.scene = SatelliteInstrumentScene2(orbit=orbit)
        self.assertEquals(self.scene.satname, "")
        self.assertEquals(self.scene.number, "")
        self.assertEquals(self.scene.orbit, orbit)
        self.assert_(self.scene.area is None)
        self.assert_(self.scene.time_slot is None)
        self.assert_(self.scene.lat is None)
        self.assert_(self.scene.lon is None)
        self.assert_(self.scene.instrument_name is None)
        self.assertEquals(self.scene.channels_to_load, set([]))
        for i, chn in enumerate(self.scene.channels):
            self.assertEquals(chn.name, channels[i][0])
            self.assertEquals(chn.wavelength_range, list(channels[i][1]))
            self.assertEquals(chn.resolution, channels[i][2])

    def test_init_time_slot(self):
        """Creation of a satellite instrument scene.
        """
        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]

        class SatelliteInstrumentScene2(SatelliteInstrumentScene):

            """Dummy satinst class.
            """
            channel_list = channels

        # time_slot

        time_slot = datetime.datetime.now()

        self.scene = SatelliteInstrumentScene2(time_slot=time_slot)
        self.assertEquals(self.scene.satname, "")
        self.assertEquals(self.scene.number, "")
        self.assertEquals(self.scene.time_slot, time_slot)
        self.assert_(self.scene.orbit is None)
        self.assert_(self.scene.area is None)
        self.assert_(self.scene.lat is None)
        self.assert_(self.scene.lon is None)
        self.assert_(self.scene.instrument_name is None)
        self.assertEquals(self.scene.channels_to_load, set([]))
        for i, chn in enumerate(self.scene.channels):
            self.assertEquals(chn.name, channels[i][0])
            self.assertEquals(chn.wavelength_range, list(channels[i][1]))
            self.assertEquals(chn.resolution, channels[i][2])

    def test_init(self):
        """Creation of a satellite instrument scene.
        """

        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]

        class SatelliteInstrumentScene2(SatelliteInstrumentScene):

            """Dummy satinst class.
            """
            channel_list = channels

        self.scene = SatelliteInstrumentScene2()
        self.assertEquals(self.scene.satname, "")
        self.assertEquals(self.scene.number, "")
        self.assert_(self.scene.time_slot is None)
        self.assert_(self.scene.orbit is None)
        self.assert_(self.scene.area is None)
        self.assert_(self.scene.lat is None)
        self.assert_(self.scene.lon is None)
        self.assert_(self.scene.instrument_name is None)
        self.assertEquals(self.scene.channels_to_load, set([]))
        for i, chn in enumerate(self.scene.channels):
            self.assertEquals(chn.name, channels[i][0])
            self.assertEquals(chn.wavelength_range, list(channels[i][1]))
            self.assertEquals(chn.resolution, channels[i][2])

    def test_setitem(self):
        """__setitem__ for sat scenes.
        """

        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]

        class SatelliteInstrumentScene2(SatelliteInstrumentScene):

            """Dummy satinst class.
            """
            channel_list = channels

        self.scene = SatelliteInstrumentScene2()

        self.assertRaises(TypeError, self.scene.__setitem__, 10.8, "rsutienrt")

        a = np.ma.array([1, 2, 3])
        self.scene[6.4] = a
        self.assertTrue(
            isinstance(self.scene[6.4].data, np.ma.core.MaskedArray))

    def test_getitem(self):
        """__getitem__ for sat scenes.
        """

        # empty scene
        self.scene = SatelliteInstrumentScene()

        self.assertRaises(KeyError, self.scene.__getitem__,
                          np.random.uniform(100))
        self.assertRaises(KeyError, self.scene.__getitem__,
                          int(np.random.uniform(10000)))
        self.assertRaises(KeyError, self.scene.__getitem__,
                          random_string(4))

        # scene with 3 channels

        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]

        class SatelliteInstrumentScene2(SatelliteInstrumentScene):

            """Dummy satinst class.
            """
            channel_list = channels

        self.scene = SatelliteInstrumentScene2()
        for chn in channels:
            self.assertEquals(self.scene[chn[0]].name, chn[0])
            for i in range(3):
                self.assertEquals(self.scene[chn[1][i]].wavelength_range[i],
                                  chn[1][i])
            self.assertEquals(self.scene[chn[2]].resolution, chn[2])
            self.assertEquals(self.scene[(chn[0], chn[2])].name, chn[0])

        self.assertRaises(KeyError, self.scene.__getitem__, [])
        self.assertRaises(KeyError, self.scene.__getitem__, random_string(5))
        self.assertRaises(TypeError, self.scene.__getitem__, set([]))
        self.assertRaises(KeyError, self.scene.__getitem__, 5.0)

        self.assertEquals(len(self.scene.__getitem__(5000, aslist=True)), 2)

        chans = self.scene.__getitem__(5000, aslist=True)
        self.assertEquals(self.scene[chans[0].name].name, channels[1][0])
        self.assertEquals(self.scene[chans[1].name].name, channels[2][0])

    def test_check_channels(self):
        """Check loaded channels.
        """

        # No data loaded

        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]

        class SatelliteInstrumentScene2(SatelliteInstrumentScene):

            """Dummy satinst class.
            """
            channel_list = channels

        self.scene = SatelliteInstrumentScene2()
        for chn in channels:
            self.assertRaises(
                NotLoadedError, self.scene.check_channels, chn[0])
            self.assertRaises(
                NotLoadedError, self.scene.check_channels, chn[2])
            for i in range(3):
                self.assertRaises(NotLoadedError,
                                  self.scene.check_channels,
                                  chn[1][i])

        # With data

        self.scene[0.7] = np.ma.array(np.random.rand(3, 3),
                                      mask=np.array(np.random.rand(3, 3) * 2,
                                                    dtype=int))
        self.scene[6.4] = np.ma.array(np.random.rand(3, 3),
                                      mask=np.array(np.random.rand(3, 3) * 2,
                                                    dtype=int))
        self.scene[11.5] = np.ma.array(np.random.rand(3, 3),
                                       mask=np.array(np.random.rand(3, 3) * 2,
                                                     dtype=int))
        self.assertTrue(self.scene.check_channels(0.7, 6.4, 11.5))
        self.assertRaises(
            KeyError, self.scene.check_channels, random_string(5))

    def test_loaded_channels(self):
        """Loaded channels list.
        """
        # No data loaded

        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]

        class SatelliteInstrumentScene2(SatelliteInstrumentScene):

            """Dummy satinst class.
            """
            channel_list = channels

        self.scene = SatelliteInstrumentScene2()
        self.assertEquals(self.scene.loaded_channels(), set([]))

        # With data

        self.scene[0.7] = np.ma.array(np.random.rand(3, 3),
                                      mask=np.array(np.random.rand(3, 3) * 2,
                                                    dtype=int))
        self.scene[6.4] = np.ma.array(np.random.rand(3, 3),
                                      mask=np.array(np.random.rand(3, 3) * 2,
                                                    dtype=int))
        self.scene[11.5] = np.ma.array(np.random.rand(3, 3),
                                       mask=np.array(np.random.rand(3, 3) * 2,
                                                     dtype=int))

        self.assertEquals(set([chn.name for chn in self.scene.loaded_channels()]),
                          set(["00_7", "06_4", "11_5"]))

    def test_project(self):
        """Projecting a scene.
        """
        area = random_string(8)
        area2 = random_string(8)
        # scene with 3 channels

        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]

        class SatelliteInstrumentScene2(SatelliteInstrumentScene):

            """Dummy satinst class.
            """
            instrument_name = random_string(8)
            channel_list = channels

        # case of a swath

        self.scene = SatelliteInstrumentScene2(area=None)

        # With data

        self.scene[0.7] = np.ma.array(np.random.rand(3, 3),
                                      mask=np.array(np.random.rand(3, 3) * 2,
                                                    dtype=int))
        self.scene[6.4] = np.ma.array(np.random.rand(3, 3),
                                      mask=np.array(np.random.rand(3, 3) * 2,
                                                    dtype=int))
        self.scene[6.4].area = MagicMock()

        res = self.scene.project(area2)
        self.assertEquals(res[0.7].shape, (3, 3))
        self.assertEquals(res[6.4].shape, (3, 3))
        self.assertRaises(KeyError, res.__getitem__, 11.5)

        self.scene[0.7].area = self.scene[6.4].area
        res = self.scene.project(area2, channels=[0.7])
        self.assertEquals(res[0.7].shape, (3, 3))
        self.assertRaises(KeyError, res.__getitem__, 6.4)

        res = self.scene.project(area2, channels=[0.7, 11.5])
        self.assertEquals(res[0.7].shape, (3, 3))
        self.assertRaises(KeyError, res.__getitem__, 11.5)

        res = self.scene.project(area2, channels=[])
        self.assertRaises(KeyError, res.__getitem__, 0.7)

        self.assertRaises(TypeError, self.scene.project,
                          area2, channels=11.5)

        # case of a grid

        self.scene = SatelliteInstrumentScene2(area=area)

        # With data

        self.scene[0.7] = np.ma.array(np.random.rand(3, 3),
                                      mask=np.array(np.random.rand(3, 3) * 2,
                                                    dtype=int))
        self.scene[6.4] = np.ma.array(np.random.rand(3, 3),
                                      mask=np.array(np.random.rand(3, 3) * 2,
                                                    dtype=int))
        self.scene[11.5] = np.ma.array(np.random.rand(3, 3),
                                       mask=np.array(np.random.rand(3, 3) * 2,
                                                     dtype=int))

        res = self.scene.project(area2)
        self.assertEquals(res[11.5].shape, (3, 3))

        self.scene[0.7].area = MagicMock()
        res = self.scene.project(area2, channels=[0.7])
        self.assertEquals(res[0.7].shape, (3, 3))
        self.assertRaises(KeyError, res.__getitem__, 6.4)

        self.scene[6.4].area = MagicMock()
        self.scene[11.5].area = MagicMock()
        res = self.scene.project(area2)
        self.assertEquals(res[0.7].shape, (3, 3))

        # case of self projection

        self.scene = SatelliteInstrumentScene2(area=area)

        # With data

        self.scene[0.7] = np.ma.array(np.random.rand(3, 3),
                                      mask=np.array(np.random.rand(3, 3) * 2,
                                                    dtype=int))
        self.scene[6.4] = np.ma.array(np.random.rand(3, 3),
                                      mask=np.array(np.random.rand(3, 3) * 2,
                                                    dtype=int))
        self.scene[11.5] = np.ma.array(np.random.rand(3, 3),
                                       mask=np.array(np.random.rand(3, 3) * 2,
                                                     dtype=int))

        self.scene[6.4].area = MagicMock()
        res = self.scene.project(area)
        self.assertEquals(res[0.7].shape, (3, 3))
        self.assertEquals(res[6.4].shape, (3, 3))
        self.assertEquals(res[11.5].shape, (3, 3))

        self.scene[11.5].area = self.scene[6.4].area
        self.scene[0.7].area = self.scene[6.4].area
        res = self.scene.project(area, channels=None)
        self.assertEquals(res[0.7].shape, (3, 3))
        self.assertEquals(res[6.4].shape, (3, 3))
        self.assertEquals(res[11.5].shape, (3, 3))

    def test_load(self):
        """Loading channels into a scene.
        """

        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]

        class SatelliteInstrumentScene2(SatelliteInstrumentScene):

            """Dummy satinst class.
            """
            instrument_name = random_string(8)
            channel_list = channels

        self.scene = SatelliteInstrumentScene2()

        self.assertRaises(TypeError, self.scene.load, "00_7")

        self.scene.load(["00_7"])
        self.assertEquals(set(), self.scene.channels_to_load)

        self.scene = SatelliteInstrumentScene2()
        self.scene.load()
        self.assertEquals(set(),
                          self.scene.channels_to_load)

        self.scene.load(["CTTH"])

        # Test the reinitialization of channels_to_load
        # self.scene = SatelliteInstrumentScene2()

        # self.assertRaises(ValueError, self.scene.load, ["00_7"], area_extent="bla")

        # self.scene.load(["00_7"], area_extent="bla")
        # self.assertEquals(set(["00_7"]), self.scene.channels_to_load)

        # self.scene.load(["06_4"])
        # self.assertEquals(len(self.scene.loaded_channels()), 1)
        # self.assertEquals(self.scene.loaded_channels()[0].name, "06_4")

        # self.scene.load(["CTTH"])

    # def test_assemble_segments(self):
    #     """Assembling segments in a single satscene object.
    #     """
    #     channels = [["00_7", (0.5, 0.7, 0.9), 2500],
    #                 ["06_4", (5.7, 6.4, 7.1), 5000],
    #                 ["11_5", (10.5, 11.5, 12.5), 5000]]

    #     class SatelliteInstrumentScene2(SatelliteInstrumentScene):
    #         """Dummy satinst class.
    #         """
    #         satname = random_string(8)
    #         number = random_string(8)
    #         instrument_name = random_string(8)
    #         channel_list = channels

    #     self.scene = SatelliteInstrumentScene2()
    #     scene2 = SatelliteInstrumentScene2()

    #     self.scene.lon = np.ma.array(np.random.rand(3, 3),
    #                                  mask = np.array(np.random.rand(3, 3) * 2,
    #                                                  dtype = int))

    #     self.scene.lat = np.ma.array(np.random.rand(3, 3),
    #                                  mask = np.array(np.random.rand(3, 3) * 2,
    #                                                  dtype = int))

    #     self.scene[0.7] = np.ma.array(np.random.rand(3, 3),
    #                                   mask = np.array(np.random.rand(3, 3) * 2,
    #                                                   dtype = int))
    #     self.scene[6.4] = np.ma.array(np.random.rand(3, 3),
    #                                   mask = np.array(np.random.rand(3, 3) * 2,
    #                                                   dtype = int))

    #     scene2.lon = np.ma.array(np.random.rand(3, 3),
    #                                  mask = np.array(np.random.rand(3, 3) * 2,
    #                                                  dtype = int))

    #     scene2.lat = np.ma.array(np.random.rand(3, 3),
    #                                  mask = np.array(np.random.rand(3, 3) * 2,
    #                                                  dtype = int))

    #     scene2[0.7] = np.ma.array(np.random.rand(3, 3),
    #                               mask = np.array(np.random.rand(3, 3) * 2,
    #                                               dtype = int))
    #     scene2[11.5] = np.ma.array(np.random.rand(3, 3),
    #                                mask = np.array(np.random.rand(3, 3) * 2,
    #                                                dtype = int))

    #     big_scene = mpop.scene.assemble_segments([self.scene, scene2])

    #     data0 = big_scene[0.7].data
    #     data1 = self.scene[0.7].data
    #     data2 = scene2[0.7].data

    #     self.assertTrue(np.ma.allclose(data0, np.ma.concatenate((data1, data2)),
    #                                    rtol=EPSILON))

    #     data0 = big_scene[0.7].data.mask
    #     data1 = self.scene[0.7].data.mask
    #     data2 = scene2[0.7].data.mask

    #     self.assertTrue(np.ma.allclose(data0, np.ma.concatenate((data1, data2)),
    #                                    rtol=EPSILON))

    #     data0 = big_scene[6.4].data
    #     data1 = self.scene[6.4].data
    #     data2 = np.ma.masked_all_like(data1)

    #     self.assertTrue(np.ma.allclose(data0, np.ma.concatenate((data1, data2)),
    #                                    rtol=EPSILON))

    #     data0 = big_scene[6.4].data.mask
    #     data1 = self.scene[6.4].data.mask
    #     data2 = data2.mask

    #     self.assertTrue(np.ma.allclose(data0, np.ma.concatenate((data1, data2)),
    #                                    rtol=EPSILON))

    #     data0 = big_scene[11.5].data
    #     data2 = scene2[11.5].data
    #     data1 = np.ma.masked_all_like(data2)

    #     self.assertTrue(np.ma.allclose(data0, np.ma.concatenate((data1, data2)),
    #                                    rtol=EPSILON))

    #     data0 = big_scene[11.5].data.mask
    #     data1 = data1.mask
    #     data2 = scene2[11.5].data.mask

    #     self.assertTrue(np.ma.allclose(data0, np.ma.concatenate((data1, data2)),
    #                                    rtol = EPSILON))

    def tearDown(self):
        """Unpatch foreign modules.
        """
        unpatch_configparser()
        unpatch_projector()


def suite():
    """The test suite for test_scene.
    """
    loader = unittest.TestLoader()
    mysuite = unittest.TestSuite()
    mysuite.addTest(loader.loadTestsFromTestCase(TestSatelliteScene))
    mysuite.addTest(loader.loadTestsFromTestCase(TestSatelliteInstrumentScene))

    return mysuite



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2012.

# SMHI,
# Folkborgsvägen 1,
# Norrköping, 
# Sweden

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Tasklist class and helper function.
"""

from mpop.saturn.filelist import FileList
from mpop import CONFIG_PATH

class TaskList(dict):
    """Defines a tasklist.
    """
    
    def __init__(self, product_file=None):
        dict.__init__(self)

        if product_file is not None:
            global_vars = {"__file__":product_file}
            local_vars = {}
            execfile(product_file, global_vars, local_vars)
            tasks = local_vars["PRODUCTS"]
            for area in tasks:
                self[area] = {}
                for product in tasks[area]:
                    self[area][product] = FileList(tasks[area][product])

    
    def split(self, *keys):
        """Split the tasklist along the *keys* parameter: keys in the first
        part, non keys in the second.
        """
        tl1 = TaskList()
        tl2 = TaskList()
        for key, item in self.items():
            if key in keys:
                tl1[key] = item
            else:
                tl2[key] = item
        return tl1, tl2

    def get_prerequisites(self, klass, area_id=None):
        """Get the channels we need to load to fulfill the tasklist according
        to methods defined in *klass*. If area is provided, account only for
        tasks on this area.
        """
        if area_id is None:
            areas = self.keys()
        elif(isinstance(area_id, (list, tuple, set))):
            areas = list(area_id)
        else:
            areas = [area_id]
        prerequisites = set()
        for area in areas:
            productlist = self.get(area, {})
            for product in productlist:
                if(hasattr(klass, product) and
                   hasattr(getattr(klass, product), "prerequisites")):
                    prerequisites |= getattr(getattr(klass, product),
                                             "prerequisites")
        return prerequisites

    def shape(self,
              klass,
              mode=set(),
              original_areas=set(),
              specific_composites=set()):
        """Shape the given the tasklist according to the options.
        """

        composites = set()

        if len(original_areas) == 0:
            original_areas = set(self.keys())

        new_tasklist = TaskList()

        if len(specific_composites) == 0 and len(mode) ==0:
            for area in original_areas:
                new_tasklist[area] = self.get(area, {})

        for i in dir(klass):
            if hasattr(getattr(klass, i), "prerequisites"):
                if("pge" in mode and
                   ("CloudType" in getattr(getattr(klass, i),
                                           "prerequisites") or
                    "CTTH" in getattr(getattr(klass, i),
                                      "prerequisites"))):
                    composites |= set([i])
                elif("rgb" in mode and
                   ("CloudType" not in getattr(getattr(klass, i),
                                               "prerequisites") or
                    "CTTH" not in getattr(getattr(klass, i),
                                          "prerequisites"))):
                    composites |= set([i])

        for area in original_areas:
            new_tasklist.setdefault(area, {})
            for product in specific_composites:
                filelist = self.get(area, {}).get(product,
                                                  [area+"_"+product+".png"])
                new_tasklist[area].setdefault(product, FileList())
                new_tasklist[area][product].extend(filelist)
            for product in composites:
                filelist = self.get(area, {}).get(product, FileList())
                if len(filelist) > 0:
                    new_tasklist[area].setdefault(product, FileList())
                    new_tasklist[area][product].extend(filelist)


        return new_tasklist

def get_product_list(satscene):
    """Returns the tasklist corresponding to the satellite described in
    *satscene*, which can be a scene object, a list or a tuple. If the
    corresponding file could not be found, the function returns more generic
    tasklists (variant and name based, then only variant based), or None if no
    file can be found.

    NB: the product files are looked for in the CONFIG_PATH directory.
    """
    
    if isinstance(satscene, (list, tuple)):
        if len(satscene) != 3:
            raise ValueError("Satscene must be a triplet (variant, name, "
                             "number) or a scene object.")
        components = satscene
    else:           
        components = [satscene.variant, satscene.satname, satscene.number]

    import os.path

    for i in range(len(components)):
        pathname = os.path.join(CONFIG_PATH,
                                "".join(components[:len(components)-i]))
        if os.path.exists(pathname+"_products.py"):
            return TaskList(pathname+"_products.py")

    return None



"""Loader for MSG, netcdf format.
"""
from ConfigParser import ConfigParser
from mpop import CONFIG_PATH
import os
import numpy.ma as ma
from numpy import array as np_array
from numpy import nan as np_nan
from glob import glob
from mpop.projector import get_area_def
import datetime 

try:
    import h5py
except ImportError:
    print "... module h5py needs to be installed"
    quit()

from mipp.xrit.MSG import _Calibrator

import logging
LOG = logging.getLogger(__name__)
#from mpop.utils import debug_on
#debug_on()

SatelliteIds = { '08': 321,  # Meteosat 8 
                  '8': 321,  # Meteosat 8 
                 '09': 322,  # Meteosat 9 
                  '9': 322,  # Meteosat 9 
                 '10': 323,  # Meteosat 10
                 '11': 324 } # Meteosat 11

channel_numbers = {"VIS006": 1,
                   "VIS008": 2,
                   "IR_016": 3,
                   "IR_039": 4,
                   "WV_062": 5,
                   "WV_073": 6,
                   "IR_087": 7,
                   "IR_097": 8,
                   "IR_108": 9,
                   "IR_120": 10,
                   "IR_134": 11,
                   "HRV": 12}

dict_channel= {'VIS006':'Channel 01','VIS008':'Channel 02','IR_016':'Channel 03','IR_039':'Channel 04','WV_062':'Channel 05','WV_073':'Channel 06',\
               'IR_087':'Channel 07','IR_097':'Channel 08','IR_108':'Channel 09','IR_120':'Channel 10','IR_134':'Channel 11','HRV':'Channel 12'}


def load(satscene, calibrate=True, area_extent=None, **kwargs):
    """Load MSG SEVIRI data from hdf5 format.
    """

    # Read config file content
    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    values = {"orbit": satscene.orbit,
    "satname": satscene.satname,
    "number": satscene.number,
    "instrument": satscene.instrument_name,
    "satellite": satscene.fullname
    }

    LOG.info("assume seviri-level4")
    print "... assume seviri-level4"

    satscene.add_to_history("hdf5 data read by mpop/msg_seviri_hdf.py")


    if "reader_level" in kwargs.keys():
        reader_level = kwargs["reader_level"]
    else:
        reader_level = "seviri-level4"

    if "RSS" in kwargs.keys():
        if kwargs["RSS"]:
            dt_end =  4
        else:
            dt_end = 12
    else:
        from my_msg_module import check_RSS
        RSS = check_RSS(satscene.sat_nr(), satscene.time_slot)
        if RSS == None:
            print "*** Error in mpop/satin/msg_seviri_hdf.py"
            print "    satellite MSG", satscene.sat_nr() ," is not active yet"
            quit()
        else:
            if RSS:
                dt_end =  4
            else:
                dt_end = 12

    print "... hdf file name is specified by observation end time"
    print "    assume ", dt_end, " min between start and end time of observation"

    # end of scan time 4 min after start 
    end_time = satscene.time_slot + datetime.timedelta(minutes=dt_end)

    filename = os.path.join( end_time.strftime(conf.get(reader_level, "dir", raw=True)),
                             end_time.strftime(conf.get(reader_level, "filename", raw=True)) % values )
    
    print "... search for file: ", filename
    filenames=glob(str(filename))
    if len(filenames) == 0:
        print "*** Error, no file found"
        return # just return without exit the program 
    elif len(filenames) > 1:
        print "*** Warning, more than 1 datafile found: ", filenames 
    filename = filenames[0]
    print("... read data from %s" % str(filename))

    # read data from hdf5 file 
    data_folder='U-MARF/MSG/Level1.5/'

    # Load data from hdf file
    with h5py.File(filename,'r') as hf:

        subset_info=hf.get(data_folder+'METADATA/SUBSET')
        for i in range(subset_info.len()):
            #print subset_info[i]['EntryName'], subset_info[i]['Value']
            if subset_info[i]['EntryName'] == "VIS_IRSouthLineSelectedRectangle":
                VIS_IRSouthLine = int(subset_info[i]['Value'])
            if subset_info[i]['EntryName'] == "VIS_IRNorthLineSelectedRectangle":
                VIS_IRNorthLine = int(subset_info[i]['Value'])
            if subset_info[i]['EntryName'] == "VIS_IREastColumnSelectedRectangle":
                VIS_IREastColumn = int(subset_info[i]['Value'])
            if subset_info[i]['EntryName'] == "VIS_IRWestColumnSelectedRectangle":
                VIS_IRWestColumn = int(subset_info[i]['Value'])
            if subset_info[i]['EntryName'] == "HRVLowerNorthLineSelectedRectangle":
                HRVLowerNorthLine = int(subset_info[i]['Value'])
            if subset_info[i]['EntryName'] == "HRVLowerSouthLineSelectedRectangle":
                HRVLowerSouthLine = int(subset_info[i]['Value'])
            if subset_info[i]['EntryName'] == "HRVLowerEastColumnSelectedRectangle":
                HRVLowerEastColumn = int(subset_info[i]['Value'])
            if subset_info[i]['EntryName'] == "HRVLowerWestColumnSelectedRectangle":
                HRVLowerWestColumn = int(subset_info[i]['Value'])
            if subset_info[i]['EntryName'] == "HRVUpperSouthLineSelectedRectangle":
                HRVUpperSouthLine = int(subset_info[i]['Value'])  # 0
            if subset_info[i]['EntryName'] == "HRVUpperNorthLineSelectedRectangle":
                HRVUpperNorthLine = int(subset_info[i]['Value'])  # 0
            if subset_info[i]['EntryName'] == "HRVUpperEastColumnSelectedRectangle":
                HRVUpperEastColumn = int(subset_info[i]['Value']) # 0
            if subset_info[i]['EntryName'] == "HRVUpperWestColumnSelectedRectangle":
                HRVUpperWestColumn = int(subset_info[i]['Value']) # 0

        sat_status=hf.get(data_folder+'METADATA/HEADER/SatelliteStatus/SatelliteStatus_DESCR')
        for i in range(subset_info.len()):
            if sat_status[i]['EntryName']=="SatelliteDefinition-NominalLongitude":
                sat_lon = sat_status[i]['Value']
                break

        #print 'VIS_IRSouthLine', VIS_IRSouthLine
        #print 'VIS_IRNorthLine', VIS_IRNorthLine
        #print 'VIS_IREastColumn', VIS_IREastColumn
        #print 'VIS_IRWestColumn', VIS_IRWestColumn
        #print 'sat_longitude', sat_lon, type(sat_lon), 'GEOS<'+'{:+06.1f}'.format(sat_lon)+'>' 

        if 1 == 0:
            # works only if all pixels are on the disk 
            from msg_pixcoord2area import msg_pixcoord2area
            print "VIS_IRNorthLine, VIS_IRWestColumn, VIS_IRSouthLine, VIS_IREastColumn: ", VIS_IRNorthLine, VIS_IRWestColumn, VIS_IRSouthLine, VIS_IREastColumn
            area_def = msg_pixcoord2area ( VIS_IRNorthLine, VIS_IRWestColumn, VIS_IRSouthLine, VIS_IREastColumn, "vis", sat_lon )
        else:
            # works also for pixels outside of the disk 
            pname = 'GEOS<'+'{:+06.1f}'.format(sat_lon)+'>'  # "GEOS<+009.5>"
            proj = {'proj': 'geos', 'a': '6378169.0', 'b': '6356583.8', 'h': '35785831.0', 'lon_0': str(sat_lon)}
            aex=(-5570248.4773392612, -5567248.074173444, 5567248.074173444, 5570248.4773392612)

            # define full disk projection 
            from pyresample.geometry import AreaDefinition
            full_disk_def = AreaDefinition('full_disk',
                                           'full_disk',
                                           pname,
                                           proj,
                                           3712,
                                           3712,
                                           aex )

            # define name and calculate area for sub-demain 
            area_name= 'MSG_'+'{:04d}'.format(VIS_IRNorthLine)+'_'+'{:04d}'.format(VIS_IRWestColumn)+'_'+'{:04d}'.format(VIS_IRSouthLine)+'_'+'{:04d}'.format(VIS_IREastColumn)
            aex = full_disk_def.get_area_extent_for_subset(3712-VIS_IRSouthLine,3712-VIS_IRWestColumn,3712-VIS_IRNorthLine,3712-VIS_IREastColumn)

            area_def = AreaDefinition(area_name,
                                      area_name,
                                      pname,
                                      proj,
                                      (VIS_IRWestColumn-VIS_IREastColumn)+1,
                                      (VIS_IRNorthLine-VIS_IRSouthLine)+1,
                                      aex )

        #print area_def
        #print "REGION:", area_def.area_id, "{"
        #print "\tNAME:\t", area_def.name
        #print "\tPCS_ID:\t", area_def.proj_id
        #print ("\tPCS_DEF:\tproj="+area_def.proj_dict['proj']+", lon_0=" + area_def.proj_dict['lon_0'] + ", a="+area_def.proj_dict['a']+", b="+area_def.proj_dict['b']+", h="+area_def.proj_dict['h'])
        #print "\tXSIZE:\t", area_def.x_size
        #print "\tYSIZE:\t", area_def.y_size
        #print "\tAREA_EXTENT:\t", area_def.area_extent
        #print "};"

        # copy area to satscene 
        satscene.area = area_def

        # write information used by mipp.xrit.MSG._Calibrator in a fake header file
        hdr = dict()

        # satellite ID number 
        hdr["SatelliteDefinition"] = dict()
        hdr["SatelliteDefinition"]["SatelliteId"] = SatelliteIds[str(satscene.sat_nr())]
        
        # processing 
        hdr["Level 1_5 ImageProduction"] = dict()
        hdr["Level 1_5 ImageProduction"]["PlannedChanProcessing"] = np_array([2,2,2,2,2,2,2,2,2,2,2,2], int)
        
        # calibration factors  
        Level15ImageCalibration = hf.get(data_folder+'METADATA/HEADER/RadiometricProcessing/Level15ImageCalibration_ARRAY')
        hdr["Level1_5ImageCalibration"] = dict()

        for chn_name in channel_numbers.keys():
            chn_nb = channel_numbers[chn_name]-1
            hdr["Level1_5ImageCalibration"][chn_nb] = dict()
            #print chn_name, chn_nb, Level15ImageCalibration[chn_nb]['Cal_Slope'], Level15ImageCalibration[chn_nb]['Cal_Offset']
            hdr["Level1_5ImageCalibration"][chn_nb]['Cal_Slope']  = Level15ImageCalibration[chn_nb]['Cal_Slope']
            hdr["Level1_5ImageCalibration"][chn_nb]['Cal_Offset'] = Level15ImageCalibration[chn_nb]['Cal_Offset']

        # loop over channels to load 
        for chn_name in satscene.channels_to_load:

            dataset_name = data_folder+'DATA/'+dict_channel[chn_name]+'/IMAGE_DATA'
            if dataset_name in hf:
                data_tmp = hf.get(data_folder+'DATA/'+dict_channel[chn_name]+'/IMAGE_DATA')

                LOG.info('hdr["SatelliteDefinition"]["SatelliteId"]: '+str(hdr["SatelliteDefinition"]["SatelliteId"]))
                #LOG.info('hdr["Level 1_5 ImageProduction"]["PlannedChanProcessing"]', hdr["Level 1_5 ImageProduction"]["PlannedChanProcessing"])
                chn_nb = channel_numbers[chn_name]-1
                LOG.info('hdr["Level1_5ImageCalibration"][chn_nb]["Cal_Slope"]:  '+str(hdr["Level1_5ImageCalibration"][chn_nb]["Cal_Slope"]))
                LOG.info('hdr["Level1_5ImageCalibration"][chn_nb]["Cal_Offset"]: '+str(hdr["Level1_5ImageCalibration"][chn_nb]["Cal_Offset"]))

                if calibrate:
                    #Calibrator = _Calibrator(hdr, chn_name)
                    bits_per_pixel = 10   ### !!! I have no idea if this is correct !!!
                    Calibrator = _Calibrator(hdr, chn_name, bits_per_pixel) ## changed call in mipp/xrit/MSG.py
                    data, calibration_unit = Calibrator (data_tmp, calibrate=1)
                else:
                    data = data_tmp
                    calibration_unit = "counts"

                LOG.info(chn_name+ " min/max: "+str(data.min())+","+str(data.max())+" "+calibration_unit )

                satscene[chn_name] = ma.asarray(data)

                satscene[chn_name].info['units'] = calibration_unit
                satscene[chn_name].info['satname'] = satscene.satname
                satscene[chn_name].info['satnumber'] = satscene.number
                satscene[chn_name].info['instrument_name'] = satscene.instrument_name
                satscene[chn_name].info['time'] = satscene.time_slot
                satscene[chn_name].info['is_calibrated'] = True

            else: 
                print "*** Warning, no data for channel "+ chn_name+ " in file "+ filename
                data = np_nan
                calibration_unit = ""
                LOG.info("*** Warning, no data for channel "+ chn_name+" in file "+filename)
                # do not append the channel chn_name




import Image
import glob
import os
from ConfigParser import ConfigParser
import numpy as np
import numpy.ma as ma
from mpop import CONFIG_PATH

import pyresample
import logging

import h5py

LOG = logging.getLogger(__name__)

ODIM_H5_FIELD_NAMES = {
   'TH': 'total_power',      # uncorrected reflectivity, horizontal
   'TV': 'total_power',      # uncorrected reflectivity, vertical
   'DBZH': 'reflectivity',    # corrected reflectivity, horizontal
   'DBZV': 'reflectivity',    # corrected reflectivity, vertical
   'ZDR': 'differential_reflectivity',    # differential reflectivity
   'RHOHV': 'cross_correlation_ratio',
   'LDR': 'linear_polarization_ratio',
   'PHIDP': 'differential_phase',
   'KDP': 'specific_differential_phase',
   'SQI': 'normalized_coherent_power',
   'SNR': 'signal_to_noise_ratio',
   'VRAD': 'velocity',
   'WRAD': 'spectrum_width',
   'QIND': 'quality_index',
   'RATE': 'precip',         # precip
   'ACRR': 'accu_precip',      # 1 hour ACCU
}


def load(satscene, *args, **kwargs):
   """Loads the *channels* into the satellite *scene*.
   """
   #
   # Dataset information
   #
   # Read config file content
   conf = ConfigParser()
   conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))

   values = {"orbit": satscene.orbit,
          "satname": satscene.satname,
          "number": satscene.number,
          "instrument": satscene.instrument_name,
          "satellite": satscene.fullname
          }

   # projection info
   projectionName = conf.get("radar-level2", "projection")
   projection = pyresample.utils.load_area(os.path.join(CONFIG_PATH, "areas.def"), projectionName)
   satscene.area = projection
   
   for chn_name in satscene.channels_to_load:
      filename = os.path.join(
         satscene.time_slot.strftime(conf.get("radar-level2", "dir", raw=True)) % values,
         satscene.time_slot.strftime(conf.get(chn_name,  "filename", raw=True)) % values )

      # Load data from the h5 file
      LOG.debug("filename: "+filename)
      filenames=glob.glob(str(filename))

      if len(filenames) == 0:
         LOG.debug("no input file found: "+filename)
         print "no input file found:"+filename
         quit()
      else:
         filename = glob.glob(str(filename))[0]
      
      # open the file
      hfile = h5py.File(filename, 'r')
      odim_object = hfile['what'].attrs['object']
      if odim_object != 'COMP':
         raise NotImplementedError('object: %s not implemented.' % (odim_object))
      else:
         # File structure
         
         #>>> hfile.keys()
         #[u'dataset1', u'dataset2', u'how', u'what', u'where']


         #>>> for f in hfile['what'].attrs.keys():
         #...  print "hfile['what'].attrs['",f,"']=", hfile['what'].attrs[f]
         #
         #hfile['what'].attrs[' object ']= COMP
         #hfile['what'].attrs[' version ']= H5rad 2.0
         #hfile['what'].attrs[' date ']= 20151201
         #hfile['what'].attrs[' time ']= 060000
         #hfile['what'].attrs[' source ']= ORG:247

         #>>> for f in hfile['where'].attrs.keys():
         #...  print "hfile['where'].attrs['",f,"']=", hfile['where'].attrs[f]
         #
         #hfile['where'].attrs[' projdef ']= +proj=laea +lat_0=55.0 +lon_0=10.0 +x_0=1950000.0 +y_0=-2100000.0 +units=m +ellps=WGS84
         #hfile['where'].attrs[' xsize ']= 1900
         #hfile['where'].attrs[' ysize ']= 2200
         #hfile['where'].attrs[' xscale ']= 2000.0
         #hfile['where'].attrs[' yscale ']= 2000.0
         #hfile['where'].attrs[' LL_lon ']= -10.4345768386
         #hfile['where'].attrs[' LL_lat ']= 31.7462153193
         #hfile['where'].attrs[' UL_lon ']= -39.5357864125
         #hfile['where'].attrs[' UL_lat ']= 67.0228327583
         #hfile['where'].attrs[' UR_lon ']= 57.8119647501
         #hfile['where'].attrs[' UR_lat ']= 67.6210371028
         #hfile['where'].attrs[' LR_lon ']= 29.4210386356
         #hfile['where'].attrs[' LR_lat ']= 31.9876502779

         # hfile['how'].attrs['nodes'] 
         # list of radar in composite

         #>>> for f in hfile['dataset1']['what'].attrs.keys():
         #...  print "hfile['dataset1'][what].attrs['",f,"']=", hfile['dataset1']['what'].attrs[f]
         #
         #hfile['dataset1'][what].attrs[' product ']= COMP
         #hfile['dataset1'][what].attrs[' startdate ']= 20151201
         #hfile['dataset1'][what].attrs[' starttime ']= 055000
         #hfile['dataset1'][what].attrs[' enddate ']= 20151201
         #hfile['dataset1'][what].attrs[' endtime ']= 060500
         #hfile['dataset1'][what].attrs[' quantity ']= RATE
         #hfile['dataset1'][what].attrs[' gain ']= 1.0
         #hfile['dataset1'][what].attrs[' offset ']= 0.0
         #hfile['dataset1'][what].attrs[' nodata ']= -9999000.0
         #hfile['dataset1'][what].attrs[' undetect ']= -8888000.0
         #>>> for f in hfile['dataset2']['what'].attrs.keys():
         #...  print "hfile['dataset2'][what].attrs['",f,"']=", hfile['dataset2']['what'].attrs[f]
         #
         #hfile['dataset2'][what].attrs[' product ']= COMP
         #hfile['dataset2'][what].attrs[' startdate ']= 20151201
         #hfile['dataset2'][what].attrs[' starttime ']= 055000
         #hfile['dataset2'][what].attrs[' enddate ']= 20151201
         #hfile['dataset2'][what].attrs[' endtime ']= 060500
         #hfile['dataset2'][what].attrs[' quantity ']= QIND
         #hfile['dataset2'][what].attrs[' gain ']= 1.0
         #hfile['dataset2'][what].attrs[' offset ']= 0.0
         #hfile['dataset2'][what].attrs[' nodata ']= -9999000.0
         #hfile['dataset2'][what].attrs[' undetect ']= -8888000.0

         _xsize = hfile['where'].attrs['xsize']
         _ysize = hfile['where'].attrs['ysize']
         #nbins= _xsize * _ysize

         #projection = hfile['where'].attrs['projdef']
         
         datasets = [k for k in hfile if k.startswith('dataset')]
         datasets.sort()
         nsweeps = len(datasets)
         
         try:
            ds1_what = hfile[datasets[0]]['what'].attrs
         except KeyError:
            # if no how group exists mock it with an empty dictionary
            ds1_what = {}
         
         _type = ''
         if 'product' in ds1_what:
            LOG.debug("product: "+ds1_what['product'])
            if ds1_what['product'] == 'COMP':
               if 'quantity' in ds1_what:
                  _type = ds1_what['quantity']
                  LOG.debug("product_type: "+_type)

                  #for chn_name in satscene.channels_to_load:
                  #   if chn_name == _type:

                  raw_data = hfile[datasets[0]]['data1']['data'][:]
                  raw_data = raw_data.reshape(_ysize,_xsize)
         
                  # flag no data
                  if 'nodata' in ds1_what:
                     nodata = ds1_what['nodata']
                     data = np.ma.masked_equal(raw_data, nodata)
                  else:
                     data = np.ma.masked_array(raw_data)
         
                  mask = np.ma.masked_array( raw_data == nodata )
                  mask = np.ma.masked_equal( mask, False)
            
                  # flag undetect data 
                  if 'undetect' in ds1_what:
                     undetect = ds1_what['undetect']
                     data[data == undetect] = np.ma.masked
                        
                  #from trollimage.image import Image as trollimage
                  #img = trollimage(mask, mode="L", fill_value=[1,1,1]) # [0,0,0] [1,1,1]
                  #from trollimage.colormap import rainbow
                  #img.colorize(rainbow)
                  #img.show()
                  #quit()

                  # gain/offset adjustment
                  if 'offset' in ds1_what:
                     offset = ds1_what['offset']
                  else:
                     offset = 0.0
                     
                  if 'gain' in ds1_what:
                     gain = ds1_what['gain']
                  else:
                     gain = 1.0

                  data *= gain + offset
                  
                  satscene[chn_name] = data
                  satscene[chn_name+'-MASK'] = mask

                  LOG.debug(" *** channel:"+chn_name)
                  
                  if _type == 'DBZH':
                     units = 'dBZ'
                  
                  if _type == 'RATE':
                     units = 'mm/h'

                  if _type == 'ACRR':
                     units = 'mm'
                     
                  satscene[chn_name].info["units"] = units
                  LOG.debug("channel:"+chn_name+" units:"+units)



#!/usr/bin/python
"""Loader for s2, jpeg2000 format.
"""
#Matias Takala FMI 2016

import glob
import os
import pickle
import re
from ConfigParser import ConfigParser

import numpy.ma as ma
from pyresample import utils

import glymur
from mpop import CONFIG_PATH
from mpop.satellites import GenericFactory

#in this version Q_V is hardcoded but could be read from metadata
QUANTIFICATION_VALUE = 10000


def parse_tile(file):
    tile = re.findall('T(\d{2}\w{3})_', file)
    f = open('s2tiles.pickle', 'r')
    s2tiles = pickle.load(f)
    f.close()
    return [tile[0], s2tiles[tile[0]]]


def read_jp2_data(file):
    jp2 = glymur.Jp2k(file)
    data = jp2[:] / (QUANTIFICATION_VALUE + 0.0)
    return data


def open_s2_tile(fname):
    data = read_jp2_data(fname)
    size = data.shape
    params = parse_tile(fname)
    areadef = utils.get_area_def(
        params[0], "Sentinel 2 tile " + params[0], 'PROJ EPSG:' + params[1][0],
        'init=epsg:' + params[1][0], size[0], size[1], params[1][1])
    return ([data, areadef])


def load(satscene):
    """Load jpeg2000 data.
    """

    # Read config file content
    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))

    for chn_name in satscene.channels_to_load:
        values = {"orbit": satscene.orbit,
                  "satname": satscene.satname.upper(),
                  "number": satscene.number,
                  "instrument": satscene.instrument_name.upper(),
                  "satellite": satscene.fullname.upper(),
                  "band": chn_name}
        filename = os.path.join(
            conf.get("msi-level2", "dir"),
            satscene.time_slot.strftime(conf.get(
                "msi-level2", "filename", raw=True)) % values)
        filelist = glob.glob(filename)
        data_area = open_s2_tile(filelist[0])
        satscene[chn_name] = ma.masked_array(data_area[0])
        satscene[chn_name].area = data_area[1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2011 SMHI

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The :mod:`mpop.plugin_base` module defines the plugin API.
"""
import weakref

class Plugin(object):
    """The base plugin class. It is not to be used as is, it has to be
    inherited by other classes.
    """
    pass

class Reader(Plugin):
    """Reader plugins. They should have a *pformat* attribute, and implement
    the *load* method. This is an abstract class to be inherited.
    """
    ptype = "reader"
    def __init__(self, scene):
        """The reader plugin takes as input a satellite scene to fill in.
        
        Arguments:
        - `scene`: the scene to fill.
        """
        Plugin.__init__(self)
        self._scene = weakref.proxy(scene)

    def load(self, channels_to_load):
        """Loads the *channels_to_load* into the scene object.
        """
        raise NotImplementedError


class Writer(Plugin):
    """Writer plugins. They must implement the *save* method. This is an
    abstract class to be inherited.
    """
    ptype = "writer"
    
    def __init__(self, scene):
        """The writer saves the *scene* to *filename*.
        
        Arguments:
        - `scene`: the scene to save.
        - `filename`: the place to save it.
        """
        Plugin.__init__(self)
        self._scene = weakref.proxy(scene)

    def save(self, filename):
        """Saves the scene to a given *filename*.
        """
        raise NotImplementedError
